;
;********************************   adcListen.asm   ************************************
;                                           by Abhinandan T. Dodamani, Widener University
;                              This file is assembled with - asm56000 -b -a -l -os,so adcListen.asm  option
;*******************  Interface to AD7703 20-bit  Analog to Digital Converter ****************
;
; This DSP56001 Assembly Program Initializes all the necessary registers to set upa communication
; link between the hardware and the DSP port. The SSI ( Synchronous Serial Interface mode is
; selected for serial link).  It transfers the sampled data to Objective -C program when called.  It does 
; the following
;	1) Enables ADC mode of operation
;	2) Selects appropriate clock frequency ( synchronous and continuous ) for data transfer
;	3) Sets word length to 24 bits
;	4) Shift direction is MSB first
;	5) Calibrates ADC, AD7703
;	6) Receives the number of data to be sampled from Objective-C
;	7) Samples the ADC to the number of data samples specified
;	8) Adjusts 4 bits, since data transfer is 24 bit length
;	9) Stores in Y space memory
;	10) Transfers sampled data to Objective-C program, when called
;
; --------  IMPORTANT: This program makes use of extra 32K RAM, to store data -------------------



; -----------------------------------------------------------------System Equates  ---------------------------------------

IPR		EQU		$FFFF	; Interrupt Priority Register
SSISR	EQU		$FFEE	; Synchronous Serial Interface(SSI) Status Register, Read only
CRA	EQU		$FFEC	; SSI Control Register A
CRB	EQU		$FFED	; SSI Control Register B
PCC		EQU		$FFE1	; Port C Control register
PCDDR	EQU		$FFE3	; Port C Data Direction register
PCD	EQU		$FFE5	; Port C Data Register
RX		EQU		$FFEF	; Serial Receive Data Register
HRX	EQU		$FFEB	; Host Receive data register
HSR	EQU		$FFE9	; Host Status register
;-------------------------------------------------------------------------------------------------------------------------------

		ORG	p:0		; Directive to specify P memory space will be the runtime memory space
RESET	JMP		>NEW	; jumps to actual program code beginning address
		
		DUP	$40-2	; sequence of source lines between DUP and ENDM directives will be 						; duplicated
		NOP			; no-operation code is duplicated
		ENDM			; ends duplication here


;----------------------------------------------  MAIN PROGRAM  -----------------------------------------------------

		ORG	P:$0040	;starting address
		
		
;-------------------------------------------------  INITIALIZE SSI PORT ----------------------------------------------

NEW	MOVEP		#$E040,X:CRA	; Initializes control register A


;Control Register A format
;_____________________________________ bit number  ____________________________________
;
;15       14       13       12      11       10       9         8         7         6         5         4         3         2        1          0
;_____________________________________   Function  _____________________________________
;
;PSR  WL1   WL0   DC4   DC3   DC2   DC1   DC0   PM7   PM6   PM5   PM4   PM3   PM2   PM1  PM0
;_____________________________________  Selection  ______________________________________
;
;  1        1        1           0        0         0        0         0         0          1         0        0         0         0         0          0   
;
;PSR -> prescaler range, set to divide by 8
;WL1, WL2 -> word length, set to 24 bits
;DC4, DC3, DC2, DC1, DC0 ->  frame rate divider control ( word transfer rate ) , set to zero
;PM7 to PM0 -> prescale modulus select (divide ratio), set to 128	



		MOVEP		#$2220,X:CRB	; Initializes control register B



;Control Register B format
;
;_____________________________________ bit number  ____________________________________
;
;  15   14    13    12     11       10       9        8        7          6            5           4          3           2        1         0
;_____________________________________   Function  _____________________________________
;
;RIE  TIE  RE  TE  MOD  GCK  SYN  FSL1  FSL0  SHFD  SCKD  SCD2  SCD1  SCD0  OF1  OF0
;_____________________________________  Selection  ______________________________________
;
; 0      0      1    0       0         0        1          0          0          0           1           0          0           0         0        0  
;
; RIE -> receive interrupt enable, disabled 
; TIE -> transmit interrupt enable, disabled
; RE -> receive enable, enabled
; TE -> transmot enable, disabled
; MOD -> SSI mode select, set to normal mode
; GCK -> gated clock control, set to continuous clock
; SYN -> sync/async control, set to synchronous
; FSL0 -> frame sync length,  RX and TX set to same length
; FSL1 -> frame sync length, RX is set to word length
; SHFD -> shift direction, set to MSB first
; SCKD -> clock source direction, set to output
; SCD2 to SCD0 -> serial control direction bits, set to input
; OF1, OF0 -> output flags, not used


		MOVEP		#$01E0,X:PCC	;initializes port C control register

;Port C Control Register format
;
;_____________________________________ bit number  ____________________________________
;
; 15   14  13  12    11   10     9          8          7          6           5           4          3          2       1         0
;_____________________________________   Function  _____________________________________
;
;  X    X   X    X     X    X      X       STD    SRD     SCK     SC2     SC1     SC0   SCLK  TXD   RXD
;_____________________________________  Selection  ______________________________________
;
;  0     0    0    0      0     0      0          1          1          1           1           0          0          0        0        0  
;
; STD ->  SSI transmit data, set to serial interface
; SRD -> SSI receive data, set to serial interface
; SCK ->  SSI serial clock, set to serial interface
; SC2 -> serial control, set to serial interface
; SC1 -> serial control, set to parallel I/O
; SC0 -> serial control, set to parallel I/O
; SCLK -> SCI serial clock, set to parallel I/O
; TXD -> SCI transmit data, set to parallel I/O
; RXD -> SCI receive data, set to parallel I/O
		

		
		MOVEP		#$001E,X:PCDDR	;initializes port C data direction register

; Port C Data Direction Register format
;
;_____________________________________ bit number  ____________________________________
;
;15   14   13   12    11   10    9         8          7          6         5         4         3           2          1          0
;_____________________________________   Function  _____________________________________
;
; X    X    X     X     X    X     X     CD8    CD7    CD6    CD5    CD4    CD3      CD2    CD1   CD0
;_____________________________________  Selection  ______________________________________
;
; 0     0     0      0     0     0     0         0         0          0         0          1          1           1         1         0  
;
; CD8 -> direction of STD, set to input 
; CD7 -> direction of SRD, set to input
; CD6 -> direction of SCK, set to input 
; CD5 -> direction of SC2, set to input 
; CD4 -> direction of SC1, set to output 
; CD3 -> direction of SC0, set to output  
; CD2 -> direction of SCLK, set to output  
; CD1 -> direction of TXD, set to output  
; CD0 -> direction of RXD, set to input


;PCD (Port C Data) format
;
; bit 0 -> RXD pin is connected to DATA READY ( active low ) pin of AD7703, A/D converter
; bit 1 -> TXD pin is connected to CHIP SELECT (active low ) ping of AD7703
; bit 2 -> SCLK selects ADC ( if low ) or DAC ( if high )
; bit 3,4 -> SC0, SC1 selects DAC #1 ( SC0 =1, SC1 = 0 ), DAC #2 ( SC0 =0, SC1 = 1 ),
;                  calibrates ADC ( SC0 = SC1 = 1 )
; bit 5 -> SC2 used for frame synchronisation, for data transmitted from ADC to the DSP port
; bit 6 -> SCK supplies clock to AD7703 and AD1862
; bit 7 -> SRD receives serial data from AD7703
; bit 8 -> STD transmits data to both AD1862, unused here in this program

;----------------------------------------------- Calibration of ADC -----------------------------------------------------

		MOVEP		#$018,X:PCD		; calibrates ADC ( SC0 = SC1 = 1 ), CHIP SELECT pin 									; Low
									; bit 2 of PCD selects ADC as mode of operation
		MOVEP		#$000,X:PCD		; End of Calibration, CHIP SELECT low
		DO			#$FFF,ECAL		;1 Second Calibration Time
		NOP
ECAL	MOVEP		#$0002,X:PCD	;Disable CHIP SELECT
		

; -------------------------------------Selects no of data to be sampled  ---------------------------------------------

		MOVE	#$2000,R1		; pointer to data buffer, stores the number of data samples in 								; $2000
		MOVE	#$2005,R0		;pointer to data buffer, stores samples data starting from $2005
		
Z1		JCLR	#0,X:$FFE9,Z1	; Waits for Host Status Register bit 0 to be set, waiting for 								; data  from Objective-C program
		MOVEP	X:HRX,Y:(R1)	; collects data from Objective-C program and stores it in $2000
		

; ------------------- Samples ADC  and stores in contiguous locations from $2005 (Y space)-------------

		.LOOP		Y:(R1)			;Samples ADC with the selected number of samples
									; Loops with the selected number of samples
RCV	JSET		#0,X:PCD,RCV	;Check DATA READY pin  = Low, if not loop back here
									
		MOVEP		#$0000,X:PCD	;Select CHIP SELECT pin = Low, enables ADC
PT1		JSET		#0,X:PCD,NXT	;wait till DATA READY pin becomes high again
									; data is being transferred to DSP port
		JMP			PT1
NXT	MOVEP		#$0002,X:PCD	;disable CHIP SELECT pin
		MOVE		X:RX,A1			; move sampled data from serial receive data register to 									; data Accumulator
		LSR 		A				; logical shift right to adjust 20 bits in 24 bit register
		LSR			A
		LSR 		A
		LSR			A
		MOVE		A1,Y:(R0)+		;moves adjusted data into Y space memory
		.ENDL						; end of loop

;----------------------- Transfers sampled data from Y space memory to   Objective -C program  ----------
		
		MOVE		#$2005,R0		; starts a loop for number of selected data samples
		.LOOP		Y:(R1)			; start of loop
PT2		JCLR		#1,X:$FFE9,PT2	; waits till bit 1 is set in the HSR
		MOVEP		Y:(R0)+,X:HRX	; transfers data from Y space memory to Host receive data 									;register
		.ENDL						; end of loop
		END		$40				; end of program
					
		