;
;**********************************   dacTalk.asm   ************************************
;                                           by Abhinandan T. Dodamani, Widener University
;                              This file is assembled with - asm56000 -b -a -l -os,so dacTalk.asm  option
;**********************  Interface to AD1862  Digital to Analog Converter  ******************
; Program reproduces standard waveforms - sine, square, pulse, ramp, triangular and DC at the set 
; frequency and amplitude, enabled from the 'Signal Generator'.
; This DSP56001 Assembly Program Initializes all the necessary registers to set upa communication
; link between the hardware and the DSP port. The SSI ( Synchronous Serial Interface mode is 
; selected for serial link).  It receives the  data from the host ( Objective -C ) program when called 
; and writes the data to the selected Digital to Analog Converter, reproducing a waveform.  The 
; program is in a continuous loop, reproducing the waveform, unless disabled from the Objective-C 
; program ( DSPClose() ).   It does the following
; 
;	1) Enables DAC mode of operation
;	2) Selects appropriate clock frequency ( synchronous and continuous ) for data transfer
;	3) Sets word length to 24 bits
;	4) Shift direction is MSB first
;	5) Resets the output of the DACs to zero volts
;	6) Selects either of the two DACs
;	7) Receives the number of data to be received from Objective-C
;	8) Receives the actual data and stores in the DSP RAM
;	9) Writes the data from RAM to DAC, reproducing the waveform
;
; --------  IMPORTANT: This program makes use of extra 32K RAM, to store data -------------------

IPR		EQU		$FFFF	; Interrupt Priority Register
SSISR	EQU		$FFEE	; Synchronous Serial Interface(SSI) Status Register, Read only
CRA	EQU		$FFEC	; SSI Control Register A
CRB	EQU		$FFED	; SSI Control Register B
PCC		EQU		$FFE1	; Port C Control register
PCDDR	EQU		$FFE3	; Port C Data Direction register
PCD	EQU		$FFE5	; Port C Data Register
TX		EQU		$FFEF	; Serial Transmit Data Register
TRX	EQU		$FFEB	; Host Transmit data register
	
		ORG	Y:$2000	; Directive to specify Y Space memory starting address

		ORG	p:0		; Directive to specify P memory space will be the runtime memory space
RESET	JMP		>NEW	; jumps to actual program code beginning address
		DUP	$40-2	; sequence of source lines between DUP and ENDM directives will be 						; duplicated
		NOP			; no-operation code is duplicated
		ENDM			; ends duplication here



;----------------------------------------------  MAIN PROGRAM  -----------------------------------------------------

		ORG		P:$0040	;starting address

		
		
;-------------------------------------------------  INITIALIZE SSI PORT ----------------------------------------------

NEW	MOVEP		#$6000,X:CRA	; Initializes control register A


;Control Register A format
;_____________________________________ bit number  ____________________________________
;
;15       14       13       12      11       10       9         8         7         6         5         4         3         2        1          0
;_____________________________________   Function  _____________________________________
;
;PSR  WL1   WL0   DC4   DC3   DC2   DC1   DC0   PM7   PM6   PM5   PM4   PM3   PM2   PM1  PM0
;_____________________________________  Selection  ______________________________________
;
;  0        1        1           0        0         0        0         0         0          0         0        0         0         0         0          0   
;
;PSR -> prescaler range, disabled 
;WL1, WL2 -> word length, set to 24 bits
;DC4, DC3, DC2, DC1, DC0 ->  frame rate divider control ( word transfer rate ) , set to zero
;PM7 to PM0 ->  set to zero, disabled


		MOVEP		#$1330,X:CRB	; Initializes control register B


;Control Register B format
;
;_____________________________________ bit number  ____________________________________
;
;  15   14    13    12     11       10       9        8        7          6            5           4          3           2        1         0
;_____________________________________   Function  _____________________________________
;
;RIE  TIE  RE  TE  MOD  GCK  SYN  FSL1  FSL0  SHFD  SCKD  SCD2  SCD1  SCD0  OF1  OF0
;_____________________________________  Selection  ______________________________________
;
; 0      0      0     1       0         0        1          1          0          0           1           1          0           0         0        0  
;
; RIE -> receive interrupt enable, disabled 
; TIE -> transmit interrupt enable, disabled
; RE -> receive enable, disabled
; TE -> transmot enable, enabled
; MOD -> SSI mode select, set to normal mode
; GCK -> gated clock control, set to continuous clock
; SYN -> sync/async control, set to synchronous
; FSL0 -> frame sync length,  RX and TX set to same length
; FSL1 -> frame sync length, RX is set to bit length
; SHFD -> shift direction, set to MSB first
; SCKD -> clock source direction, set to output
; SCD2  -> serial control direction bit, set to output
; SCD0 , SDC1 -> serial control direction bits, set to input
;OF1, OF0 -> output flags, not used

		
		MOVEP		#$1E0,X:PCC

;Port C Control Register format
;
;_____________________________________ bit number  ____________________________________
;
; 15   14  13  12    11   10     9          8          7          6           5           4          3          2       1         0
;_____________________________________   Function  _____________________________________
;
;  X    X   X    X     X    X      X       STD    SRD     SCK     SC2     SC1     SC0   SCLK  TXD   RXD
;_____________________________________  Selection  ______________________________________
;
;  0     0    0    0      0     0      0          1          1          1           1           0          0          0        0        0  
;
; STD ->  SSI transmit data, set to serial interface
; SRD -> SSI receive data, set to serial interface
; SCK ->  SSI serial clock, set to serial interface
; SC2 -> serial control, set to serial interface
; SC1 -> serial control, set to parallel I/O
; SC0 -> serial control, set to parallel I/O
; SCLK -> SCI serial clock, set to parallel I/O
; TXD -> SCI transmit data, set to parallel I/O
; RXD -> SCI receive data, set to parallel I/O


		MOVEP		#$01E,X:PCDDR

; Port C Data Direction Register format
;
;_____________________________________ bit number  ____________________________________
;
;15   14   13   12    11   10    9         8          7          6         5         4         3           2          1          0
;_____________________________________   Function  _____________________________________
;
; X    X    X     X     X    X     X     CD8    CD7    CD6    CD5    CD4    CD3      CD2    CD1   CD0
;_____________________________________  Selection  ______________________________________
;
; 0     0     0      0     0     0     0         0         0          0         0          1          1           1         1         0  
;
; CD8 -> direction of STD, set to input 
; CD7 -> direction of SRD, set to input
; CD6 -> direction of SCK, set to input 
; CD5 -> direction of SC2, set to input 
; CD4 -> direction of SC1, set to output 
; CD3 -> direction of SC0, set to output  
; CD2 -> direction of SCLK, set to output  
; CD1 -> direction of TXD, set to output  
; CD0 -> direction of RXD, set to input


;PCD (Port C Data) format
;
; bit 0 -> RXD pin is connected to DATA READY ( active low ) pin of AD7703, A/D converter
; bit 1 -> TXD pin is connected to CHIP SELECT (active low ) ping of AD7703
; bit 2 -> SCLK selects ADC ( if low ) or DAC ( if high )
; bit 3,4 -> SC0, SC1 selects DAC #1 ( SC0 =1, SC1 = 0 ), DAC #2 ( SC0 =0, SC1 = 1 ),
; bit 5 -> SC2 used for frame synchronisation, for data transmitted from DSP port to DAC
; bit 6 -> SCK supplies clock to AD7703 and AD1862
; bit 7 -> SRD receives serial data from AD7703
; bit 8 -> STD transmits data to both AD1862, unused here in this program

;------------------ Transmit  Initialization for both DACs  - Reset output to Zero Volts -------------------

		MOVEP		#$00E,X:PCD			; DAC 1 (AD 1862 ) selection
		MOVEP		#$000000,X:TX		; 20 -bit data, resets output to zero Volts
		NOP							; time delay
		NOP							; time delay
		MOVEP		#$016,X:PCD			;DAC 2 ( AD 1862 ) selection



;--------------------------------------------  Selects DAC from Object-C  ------------------------------------------

LOOP1	JCLR	#0,X:$FFE9,LOOP1	; Waits for Host Status Register bit 0 to be set, waiting
									; for data which selects DAC #1 or DAC #2
		MOVEP	X:TRX,X:PCD		;Selects DAC from Host
		

;---------------------------------- Selects number of data to collect from Objective-C----------------------------

		MOVE	#$2000,R1			; pointer to data buffer, stores the number of data samples 									; in $2000
LOOP2	JCLR	#0,X:$FFE9,LOOP2	; Waits for Host Status Register bit 0 to be set, waiting for 									; data  from Objective-C program
		MOVEP	X:TRX,Y:(R1)		; Register R1 contains number of data to collect from host

		
;----------------------------------------  Transfer Data from  Host ( Object-C )   --------------------------------
		
		
		
		MOVE		#$2005,R0			; pointer to data buffer, stores samples data starting  
										; from$2005
		.LOOP		Y:(R1)				; Loops and reads data , the number of data already  
										; set in R1 - pointed to location  $2000
CURL	JCLR		#0,X:$FFE9,CURL	; waits for data from host and stores in location 
		MOVEP		X:TRX,Y:(R0)+		; pointed by R0, $2005
		NOP							; no operation, delay
		.ENDL							; end of loop


;-------------------------------------------------   Writes  data to DAC   ---------------------------------------------

HERE	MOVE		#$2005,R0			; resets pointer to start of data buffer
		.LOOP		Y:(R1)				; sets a loop with number of data
		MOVEP		Y:(R0)+,X:TX		; writes data to selected DAC
		DO			#935,NOW			; Loop to adjust frequency
		NOP							; delay
NOW	NOP							; delay
		.ENDL							; end of loop
AAA	JMP			HERE				; starts the routine again
		END		$40					; end of program
					
		