#include <libc.h>
#include "filter.h"

void cxRatfnError(char *s=0, int i=-1) {
    cerr << "Error: complex ratfn used to create a real filter. ";
    if (s)
        cerr << s ;
    if (i!=-1)
	cerr << " (" << i << ")";
    cerr << "\n";
    exit(-1);
}

filter::filter(ratfn &h) { 
    if (h.den.degree() < h.num.degree()) {
	cerr << "Error: Attempt to define a filter with an improper transfer function.\n";
	exit( -1 );
    }
    numSections = h.num.degree()/2;	// No. of 2nd-order numerator sections
    numIsOdd = h.num.degree() & 1;	// Is there a 1st-order section?
    denSections = h.den.degree()/2;
    denIsOdd = h.den.degree() & 1;

    state.set_size(h.den.degree());
    state.initialize();

    // Determine the feedforward coefficients
    h.num.sort_roots();
    h.den.sort_roots();
    feedforward.set_size(h.num.degree()+1);
    complex z = h.num.lcoeff/h.den.lcoeff; 
    if (!isReal(z))
        cxRatfnError("The leading coefficient is complex.");
    feedforward[0] = z.re;
    double *ff = &feedforward[0]+1;
    complex *root = &h.num.root[0];

    if (!isReal(z)) cxRatfnError();
    if (numIsOdd) {
        z = *root++;
        if( !isReal(z) )
            cxRatfnError("the numerator order is odd, but there are no real roots");
        *ff++ = -z.re;
    }
    for (int i=0; i<numSections; ++i) {
         z = *root + *(root+1);
         // isReal(z) doesn't do the trick for (eps1+j, eps2-j)
         if (!isReal(z, 1e-6) && fabs(z.im) > 1e-6*(fabs(root->im)+fabs((root+1)->im))) {
             cerr.precision(10);
	     cerr << *root << ", " << *(root+1) << endl;
             cxRatfnError("non-opposite im. parts in the numerator roots", i);
	 }
         *ff++ = -z.re;
         z = *root * *(root+1);
         if (!isReal(z, 1e-6)) {
  	     cerr.precision(10);
	     cerr << *root << ", " << *(root+1) << endl;
             cxRatfnError("non-opposite phases in the numerator roots", i);
	 }
         *ff++ = z.re;
         root += 2;
    }

// Determine the feedback coefficients
    // CAN DO OPTIMIZATIONS FOR FIR FILTERS???
    feedback.set_size(h.den.degree());
    double *fb = &feedback[0];
    root = &h.den.root[0];
    if (denIsOdd) {
        z = *root++;
        if (!isReal(z)) cxRatfnError();
        *fb++ = z.re;
    }
    for (i=0; i<denSections; ++i) {
         z = *root + *(root+1);
         if (!isReal(z))
             cxRatfnError("non-opposite im. parts in the denominator roots, i");
         *fb++ = z.re;
         z = *root * *(root+1);
         if (!isReal(z))
             cxRatfnError("non-opposite phases in the denominator roots", i);
         *fb++ = -z.re;
         root += 2;
    }
}

// This is rather messy.
// Perhaps more flags or a better structure would clean it up.
double filter::operator()(double u) {
    register double *ff=&feedforward[0], *lastff=ff+feedforward.size(),
      *fb=&feedback[0];
    register double x0, *x1=&state[0], *x2=x1+1;
    double y = u * *ff++;
    if (denIsOdd) {	// Special case: the first section is first-order.
	x0 = y + *fb++ * *x1;
	if (numIsOdd)	// Pair the real den. root with a real num. root.
	    y = x0 + *ff++ * *x1;
	else 
	    y = *x1;
	*x1++ = x0;
	x2++;
    }
    for (int i=0; i<denSections; ++i,x1+=2,x2+=2) {
  	 x0 = y + *fb++ * *x1;
	 x0 += *fb++ * *x2;
	 if (i==0 && numIsOdd && !denIsOdd) // first-order numerator
	     y = *x1 + *ff++ * *x2;
	 else if( ff < lastff )  // second-order numerator
	     y = x0 + *ff++ * *x1 + *ff++ * *x2;
	 else	//zeroth-order numerator
	     y = *x2;
	 *x2 = *x1;
	 *x1 = x0;
    }
    return y;
}

void filter::reset() {
    state.initialize();
}

cxfilter::cxfilter(ratfn &h) { 
    order=h.den.degree();
    if (order < h.num.degree()) {
	cerr << "Attempt to define a filter with an improper transfer function.\n";
	exit( -1 );
    }

    feedforward.set_size(order+1);
    complex a_n = h.den.coefficient(order);
    for (int i=0; i<feedforward.size(); ++i)
	 feedforward[i] = h.num.coefficient(order-i)/a_n;

    // only make the feedback vector as large as necessary (FIR filters esp)
    for (i=0; i<order, abs(h.den.coefficient(i)) < 100*abs(a_n)*macheps(); ++i)
	;
    feedback.set_size(order-i);
    for (i=0; i<feedback.size(); ++i)
 	 feedback[i] = -h.den.coefficient(order-i-1)/a_n;

    t=0;
    delay_chain.set_size(order);		// implicitly initialized to 0
}

complex cxfilter::simulate(complex u) {
    complex x, y;
    x = u;
    for (int i=0; i<feedback.size(); ++i)
	 x += feedback[i] * delay(i+1);
    y = x * feedforward[0];
    for (i=1; i<feedforward.size(); ++i)
	 y += feedforward[i] * delay(i);
    shift_in(x);
    return y;
}

// put x into the delay_chain vector
void cxfilter::shift_in(const complex x) {
    if (order == 0)
	return;
    if (t)
	t--;
    else
	t=order-1;
    delay_chain[t]=x;
}

void cxfilter::reset() {
    for (int i=0; i<order; ++i) 
	 delay_chain[i]=0;
    t=0;
}

ostream& operator<<(ostream& s, const filter& f) {
    s << "feedback coefficients= " << f.feedback << endl;
    s << "feedforward coefficients= " << f.feedforward << endl;
    s << "state vector= " << f.state << endl;
    return s;
}
