#include <libc.h>
#include <stdlib.h>  /* for qsort() */
#include "polynomial.h"
#include <numerics.h>

#define LBRACE '{'
#define RBRACE '}'

char 	*polynomial::coeffPrefix="coeffs=",  // initialize the static vars
	*polynomial::rootPrefix="roots=",
	*polynomial::lcoeffPrefix="lcoeff=";

polynomial::polynomial(const complex x) { 
	n=0;
	coeff.set_size(1);
	root.set_size(0);
	coeff[0]=lcoeff=x;
	valid.coeffs=1;
	valid.roots=1;
}

polynomial& polynomial::operator=( const polynomial& p ) {
	lcoeff = p.lcoeff;
	root = p.root;
	n = p.n;
	coeff = p.coeff;
	valid = p.valid;
	return *this;
}

void polynomial::set_roots( const complex a, const cxVector& v ) {
    lcoeff = a;
    root = v;
    n = v.size();
    coeff.set_size(n+1);
    valid.coeffs=0;
    valid.roots =1;
}

void polynomial::set_coeffs( const cxVector& v ) {
    coeff = v;
    n = v.size()-1;
    root.set_size(n);
    valid.coeffs=1;
    valid.roots =0;
}

void polynomial::bless_coeffs() {
    if(!valid.coeffs) 
	calculate_coeffs();
}

int relativeSize( double x, double y ){
    if( fabs(x-y) <= 1e-10*(fabs(x)+fabs(y)) )
	return 0;
    else
	return x>y?1:-1;
}

// return 1 if fabs(im(z1)) is significantly greater than that of z2 
// or if their real parts are similarly related 
// when the abs values of the imaginary parts are comparable.
int cxCompare( const void *z1, const void *z2 ) {
    int cmp;
    if( (cmp=relativeSize(fabs(((complex*)z1)->im),fabs(((complex*)z2)->im)))==0 ) // imaginary parts are comparable
	return relativeSize(((complex*)z1)->re,((complex*)z2)->re);
    else 
	return cmp;
}

void polynomial::sort_roots() {
    bless_roots();
    qsort( &root[0], root.size(), sizeof(complex), cxCompare );
    // Ensure that the roots are ordered in complex-conjugate pairs.
    // The danger is that repeated complex roots are not thus arranged.
    for( int i=0; i<root.size(); ++i ) {
	if( isReal(root[i]) )
	    continue;
	if( relativeSize(root[i].im, -root[i+1].im) == 0 ) { 
	    // Just ensure that the one with the positive im. part is first.
	    if( root[i].im <0 ){
		 complex tmp=root[i];
		 root[i] = root[i+1];
		 root[i+1] = tmp;
		 }
	    ++i;	// go to the next pair
	}
	else {// uh-oh: need to find the matching root
	    for( int j=i+2; j<root.size() && 
	         cxCompare(&root[i],&root[j])==0; ++j ) {
		if( relativeSize(root[i].im, -root[j].im) == 0 ) {
		    complex tmp=root[++i];
		    root[i] = root[j];
		    root[j] = tmp;
		    // Ensure that the one with the positive im. part is first.
 		    if( root[i-1].im <0 ) {
			complex tmp=root[i];
			root[i] = root[i-1];
			root[i-1] = tmp;
		    }
		}
            }
        }
    } // for(...)
}

void polynomial::make_consistent() {
    // ensure consistency between the redundant representations of a polynomial
    if( !(valid.roots || valid.coeffs) ) {
	cerr << "Polynomial error: polynomial completely undefined.\n";
	exit( -1 );
    }
    bless_coeffs();
    bless_roots();
}

void polynomial::change_degree(const int d) {
    if(n!=d) {
       if( valid.roots )
           root.change_size(n=d);
       else
           root.set_size(n=d);
       if( valid.coeffs )
           coeff.change_size(n+1);
       else
           coeff.set_size(n+1);
    }
}

// remove zeros in the high-order coefficients
void polynomial::verify_degree() {
    if( !valid.coeffs )
	return;
    for( int m=n; m>0 && (abs(coeff[m]) < 10*n*macheps()); --m );
    change_degree(m);
}

void polynomial::calculate_roots() {
    if( !valid.coeffs ) {
	cerr << "Attempt to calculate polynomial roots when coefficients invalid.\n";
	exit(-1);
    }
    verify_degree();
    polynomial p;
    p.set_coeffs(coeff);	// holder for the deflated polynomial
    lcoeff=coeff[n];
    complex r=0;
    for( int i=0; i<n; ++i ) {
        p.laguerre(r);		 // could use quadratic formula if i==n-2
	if( abs(imag(r))< 10*macheps()*abs(real(r)) )
	    r = real(r);
	p.deflate(r);
	root[i] = r;
	r = conj(r);
    }
    polish_roots();
    valid.roots =1;
}

// Use Laguerre's method to find a root.
// See section 9.5 of "Numerical Recipes in C"
// by W.H. Press, B.P. Flannery, S.A. Teukolsky and W.T. Vetterling 
// Cambridge University Press, 1988
void polynomial::laguerre( complex &x ) const {
    complex p0, p1, p2;
    complex g, g2, h, surd, dp, dm, a;
    double dx, last_dx=0;
    enum { no, ftol, xtol, diverge, itmax } done=no;

    for( int iteration=1; !done; ++iteration ) {
	evaluate_derivatives012( x, p0, p1, p2 );
	if( abs(p0) < macheps() * (abs(p1)+1) ) {
	    // done=ftol;
	    break;
        }
	g=p1/p0; g2=g*g; h=g2-p2/p0;
	surd = sqrt( (n-1)*(n*h-g2) );
	dp=g+surd; dm=g-surd;
	if( norm(dp) > norm(dm) )
	    a = n/dp;
	else
	    a = n/dm;

	// stopping criteria
	dx=abs(a);
	if( iteration > 15 && dx >= last_dx )
	    done = diverge;
	else {
	    x -= a;
	    if( iteration > 100 )
		done = itmax;
	    else if( dx <= 100*macheps()*abs(x) )
		done = xtol;
        }
	last_dx = dx;
    }
    // cerr << "laguerre root= "<< x <<", done="<< done << endl;
}

// evaluate the polynomial, its first derivative and and its second derivative
void polynomial::evaluate_derivatives012( 
  const complex x, complex &p0, complex &p1, complex &p2 ) const {
    p0=coeff[n]; p1=0; p2=0;
    for( int i=n-1; i>=0; --i ) { // calculate p, p', p'' efficiently
	p2 = x*p2 + p1;
	p1 = x*p1 + p0;
	p0 = x*p0 + coeff[i];
    }
}

complex polynomial::deflate( const complex a ) {
    complex rem=coeff[n];
    for( int i=n-1; i>=0; --i ) {
	complex tmp=coeff[i];
	coeff[i]=rem;
	rem=tmp+rem*a;
    }
    if( n )
        coeff.change_size(n--); // somewhat a waste of time
    valid.roots=0;
    return rem;
}

void polynomial::polish_roots() { /* Use Newton's method */
    complex p0, p1;
    for( int i=0; i<n; ++i ) {
	complex r=root[i];
	double big_change = 0.1*(1+abs(r));
	for( int itn=0; itn<5; ++itn ) {
	    evaluate_derivatives01( r, p0, p1 );
	    if( abs(p0) >= big_change * abs(p1) )
		break;
	    r -= p0/p1;
        }
	if( abs(r-root[i]) < .5 * big_change )
	    /* polishing hasn't altered the root much */
	    root[i]=r;
    }
}

// evaluate the polynomial and its first derivative from the coefficients
void polynomial::evaluate_derivatives01( 
  const complex x, complex &p0, complex &p1 ) const {
    p0=coeff[n]; p1=0;
    for( int i=n-1; i>=0; --i ){ 	// calculate p, p', efficiently
	p1 = x*p1 + p0;
	p0 = x*p0 + coeff[i];
    }
}
void polynomial::calculate_coeffs() {
    if( !valid.roots ) {
	cerr << "Attempt to calculate polynomial coefficients when roots invalid.\n";
	exit(-1);
    }
    int i;
    coeff[0]=lcoeff;
    for( i=0; i<n; ++i ) {
	const complex r=root[i];
	int j=i+1; 
	register complex *cj=&coeff[j], *cjm1=&coeff[j-1];
	*cj = *cjm1;
	while( --j>0 ) {
	    cj = cjm1--;
	    *cj = *cjm1 - (*cj)*r;
        }
	*cjm1 *= -r;
    }
    valid.coeffs=1;
}

complex polynomial::operator() ( complex x ) {
    if( valid.roots ) {	// prefer the root representation for evaluation
	complex p=lcoeff;
	if(n>0) {
	    complex *r=&root[0];
	    int i=n;
	    while( i-- )
		p *= x - *r++;
        }
	return p;
    }
    else if( valid.coeffs ) {
	complex *c=&coeff[n];
	complex p=*c;
	int i=n;
	while( i-- )
	    p = x * p + *--c;
	return p;
    }
    else {
	cerr << "Attempted evaluation of an invalid polynomial.\n";
	exit( -1 );
	return 0;
    }
}

ostream& operator<<( ostream& s, const polynomial &p ) {
    if( p.valid.roots )
	s << polynomial::lcoeffPrefix << ' ' << p.lcoeff
	  << ' ' << polynomial::rootPrefix << ' ' << p.root;
    else if( p.valid.coeffs ) {
	s << polynomial::coeffPrefix << ' ' << LBRACE << ' ';
        for( int i=p.coeff.size(); i-->0; )
	    s << p.coeff[i] << " ";
        s << RBRACE;
    }
    else
        s << "Undefined polynomial";
    return s;
}

#include <string.h>
istream& operator>>( istream& s, polynomial& p ) {
    char ident[128];  //!! YEUCH
    cxVector v;
    s >> ident;
    if (strcmp(ident,polynomial::coeffPrefix)==0) { // for coeff representation
        s >> v;
        v.reverse();
        p.set_coeffs( v );
    }
    else { // root representation
        complex lc;
	if(strcmp(ident,polynomial::rootPrefix)==0) {
	    s >> v >> ident;
	    if(strcmp(ident,polynomial::lcoeffPrefix)==0)
		s >> lc;
	    else
		// should put ident back and assume lc=1
		cerr << "missing " << polynomial::lcoeffPrefix << endl;
	}
	else if(strcmp(ident,polynomial::lcoeffPrefix)==0) {
	    s >> lc >> ident;
	    if(strcmp(ident,polynomial::rootPrefix)==0)
		s >> v;
	    else
		// should put ident back
		cerr << "missing " << polynomial::rootPrefix << endl;
        }
	else
	    cerr << "Error in specifiying a polynomial. Check the prefixes.\n";
	p.set_roots( lc, v );
    }
    // p.make_consistent();
    return s;
}

polynomial operator+( polynomial& p1, polynomial& p2 ) {
    if( p1.degree() < p2.degree () )
        return(p2+p1);
    // use the coefficient representation for doing addition
    p1.bless_coeffs();
    p2.bless_coeffs();
    polynomial p;
    p = p1;
    p.valid.roots = 0; // the roots get changed after addition.
    for(int i=0; i<=p2.degree(); ++i)
	p.coeff[i] += p2.coeff[i];
    p.verify_degree();
    // could find the roots,
    // polish them using p1+p2 and
    // re-calculate coefficients
    return p;
}

polynomial operator+(polynomial& p1, const complex& c) {
    //add the complex number to the leading coefficient of the poly.
    polynomial p;
    if (p1.valid.coeffs == 0) {
        //coefficients are not present. calculate them.
        p1.bless_coeffs();
    }
    p = p1; //copy p1 to p and add c
    p.coeff[0] += c;
    p.valid.roots = 0; //as roots have changed now.
    return (p);
}

polynomial operator-( const polynomial& p ) {
    polynomial pm;
    pm=p;
    if( p.valid.coeffs ) {
	for(int i=0; i<=p.degree(); ++i)
	    pm.coeff[i] = -p.coeff[i];
    }
    if( p.valid.roots )
	pm.lcoeff = -p.lcoeff;
    return pm;
}

polynomial operator-(transform_variable, complex& c) {
    polynomial p;
    p.set_degree(1);
    p.coeff[0] = c;
    p.coeff[1] = 1;
    p.valid.coeffs = 1;
    return (p);
}

polynomial operator-(transform_variable, int n) {
    polynomial p;
    p.set_degree(1);
    p.coeff[0] = n;
    p.coeff[1] = 1;
    p.valid.coeffs = 1;
    p.valid.roots = 0;
    return (p);
}

polynomial operator-(polynomial& p1, transform_variable) {
    polynomial p;
    if (p1.valid.coeffs == 0) {//find the coefficients
        p1.bless_coeffs();
    }
    p = p1; //as the degree will not change
    p.coeff[1] = 1; //for the transform_variable 'z'
    p.valid.roots = 0;
    return(p);
}

// Unfortunately, ^ has an inappropriate precedence (lower than * and +).
polynomial operator^( transform_variable, int n ) {
    polynomial p;
    p.set_degree(n);
    p.coeff[n]=1;
    p.valid.coeffs=1;
    return p;
}

polynomial operator*( complex z, polynomial& p1 ) {
    polynomial p=p1;
    // use the root representation for doing multiplication
    if( p.valid.roots )
	p.lcoeff *= z;
    if( p.valid.coeffs )
	for( int i=0; i<=p.n; ++i )
	    p.coeff[i] *= z;
    return p;
}

polynomial operator*( polynomial& p1, polynomial& p2 ) {
    polynomial p;
    // use the root representation for doing multiplication
    p1.bless_roots();
    p2.bless_roots();
    p.set_roots( p1.lcoeff * p2.lcoeff, cxVector(p1.root, p2.root) );
    return p;
}

polynomial operator*( transform_variable, polynomial p ) {
    p.add_zero_roots(1);
    return p;
}

polynomial divide( polynomial& p1, polynomial& p2 ) {
    polynomial p;
    // use the coefficient representation to do division.
    // p1 contains the remainder
    int degree_diff=p1.degree()-p2.degree();
    if( degree_diff < 0 )
	return 0;
    p.set_degree(degree_diff);
    p1.bless_coeffs();
    p2.bless_coeffs();
    for( int i=degree_diff, j=p1.degree(); i>=0; --i ) {

    /*!!! trying to get rid of an error message
	complex q = p.coeff[i] = p1.coeff[j--]/p2.coeff[p2.degree()]; 
    !!!! substitute this instead */ 
	complex q=p1.coeff[j--]/p2.coeff[p2.degree()];
	p.coeff[i]=q;

	for( int k1=j, k2=p2.degree()-1; k2>=0; )
	     p1.coeff[k1--] -= q*p2.coeff[k2--];
    }
    p.valid.coeffs = 1;
    p1.valid.roots = 0;
    p1.change_degree(p2.n-1);
    return p;
}

void polynomial::remove_root( int i ) {
    for( int j=i; j<n-1; ++j )
	root[j]=root[j+1];
    valid.coeffs=0;
    change_degree(n-1);
}

void polynomial::add_zero_roots( int i ) {
    if( i==0 )
	return;
    change_degree(n+i);
    if( valid.coeffs ) {	// pad the coeff[] vector with zeros
	for( int j=n; j>=i; --j )
	    coeff[j] = coeff[j-i];
	while( j >= 0 )
	    coeff[j--] = 0;
    }
    if( valid.roots )		// add zeros to the root[] vector
	while( i > 0 )		// this is unnecessary if new creates zeros
	    root[n-i--] = 0;
}

polynomial invert_argument( const polynomial& p1 ) {	
    polynomial p;
    int n=p1.degree();

    p.set_degree(n);
    if( p1.valid.coeffs ) {
	// for the coefficient representation, just copy them in reverse order
	for( int i=0; i<=n; ++i )
	    p.coeff[i]= p1.coeff[n-i];
	p.valid.coeffs = 1;
	p.verify_degree();
    }
    if( p1.valid.roots ) {
	// for the root representation, take the inverse of the non-zero roots
	// and multiply the leading coefficient by the product of the roots
	p.lcoeff = p1.lcoeff;
	for( int i=0,j=0; j<n; ++j ) {
	    if( abs(p1.root[j]) > 10*n*macheps() ) {
		if( i> p.degree() ) { // doesn't jive with coeff[] results
		    cerr << "polynomial error. Too many non-zero roots.\n";
		    break;
                }
		p.root[i++]= 1/p1.root[j];
		p.lcoeff *= -p1.root[j];
	    }
        }
	p.valid.roots = 1;
	// must correct the degree if there was a zero root and we couldn't 
	// have noticed it by doing the coefficients
	if( !p.valid.coeffs )  
	    p.change_degree(i);
    }
    return p;
}

polynomial integral( polynomial& p1 ) {
    polynomial p;
    int n=p1.degree();

    p.set_degree(n+1);
    p1.bless_coeffs();
    for( int i=n+1; i>0; --i )
	p.coeff[i]= p1.coeff[i-1]/i;
    p.coeff[0] = 0;	// not necessary if new sets stuff to zero
    p.valid.coeffs = 1;
    return p;
}

int polynomial::multiplicity( int i ) {
    // this code is a mess. 
    // Perhaps it is time to consider improving the root representation...
    // return the multiplicity of root i
    complex r;
    bless_roots();
    if( valid.roots != 2 ) {
	// Group equal roots together.  THIS CHANGES THE MEANING OF i!!!
	// (but i is 0 in this case)
	for( int j=0; j<n-1; ++j ) {
	    r=root[j];
	    for( int k=j+1; k<n; ++k ) { 
		// search the root vector for equal roots
		if( norm(root[k]-r) >= 100*macheps() )
		    continue;
		// place equal roots side-by-side in the root vector
		complex tmp=root[k];
		++j;
		for( int l=k; l>j; --l )
		    root[l]=root[l-1];
		root[j]=tmp;	
	    }
        }
	valid.roots = 2;
    }
    // count multiplicity.
    r=root[i];
    for( int m=1, j=i+1; j<n; ++j ) {
	if( norm(root[j]-r) < 100*macheps() )
	    ++m;
	else
	    break;
    }
    for( j=i-1; j>=0; --j ) {
	if( norm(root[j]-r) < 100*macheps() )
	    ++m;
	else
	    break;
    }
    return m;
}

complex polynomial::reduced_evaluation( int i ) {
    // calculate poly(z)/(z-root[i])^m at z=root[i];
    bless_roots();
    complex z=root[i], p=lcoeff;
    if(n>0) {
	complex *r=&root[0];
	int i=n;
	while( i-- ) {
	    complex diff= z - *r++;
	    if( norm(diff) > 100*macheps() )
	        p *= diff;
	}
    }
    return p;
}

polynomial polynomial::reduced_polynomial( int i ) {
    // calculate poly(z)/(z-root[i])^m;
    polynomial p;
    int m=multiplicity(i);

    p.set_degree(n-m);
    p.lcoeff = lcoeff;
    complex *p_rootPtr = &p.root[0];
    for( int j=0; j<i; ++j )
	*p_rootPtr++ = root[j];
    for( j=i+m; j<n; ++j )
	*p_rootPtr++ = root[j];

    p.valid.roots = 2; // roots are grouped as a side-effect of multiplicity()
    return p;
}
