#include "ratfn.h"
#include <fstream.h>

double choose(int, int);
double factorial(int);

complex partial_fraction::operator()(complex z) {
    z -= root;
    return coeff/pow(z,multiplicity);
}

complex partial_fraction::evaluate_integral(complex z1, complex z2) {
    z1 -= root; z2 -= root;
    if( multiplicity == 1 ) // this will only work for short contours
	return coeff*log( z2/z1 );
    else {
	int m1=multiplicity-1;
	return coeff/m1 * (1/pow(z1,m1) - 1/pow(z2,m1));
    }
}
/* THIS ROUTINE DOES NOT TAKE CARE OF FOLLOWING CASES :
   IF n=m=0, IT SHOULD RETURN 1. HERE IT RETURNS (0/0)!!!
   IF m=1, THE WHILE LOOP DOES NOT EXECUTE EVEN ONCE. THIS
   WILL NOT FIND n! AT ALL. AND IT CALCULATES ONLY n!/m!
double choose( int n, int m ){   calculate n! / m!(n-m)!
    double num=n,den=m;
    while( m>1 ){
	num *= --n;
	den *= --m;
    }
    return num/den;   actually an int, but I'll use double to forestall overflow
}*/
// MODIFIED ROUTINE FOR choose
double choose(int n, int m) {
    double num, den;
    num = factorial(n);
    den = factorial(m) * factorial(n-m);
    return(num/den);
}

double factorial(int n) {
    double f = n;
    if ((n==0) || (n==1))
         return(1);
    while (n>1) {
         f *= --n;
    }
    return(f);
}

complex partial_fraction::invert_z_transform(int n) {
    // Assume the region of convergence includes the unit circle
    if( n < multiplicity )
	return 0;
    switch( multiplicity ) {
	case 1:
	    return coeff * pow(root,n-1);
        case 2:
	    return (n-1) * coeff * pow(root,n-2);
	default:
	    return choose(n-1,multiplicity-1) * coeff * pow(root,n-multiplicity);
    }
}

ratfn::ratfn(const partial_fraction pf ) {
    num = pf.coeff;
    num.valid.coeffs=1;
    den.set_degree(pf.multiplicity);
    den.lcoeff = 1;
    for( int i=0; i<pf.multiplicity; ++i )
	den.root[i] = pf.root;
    den.valid.roots=1;
}

// taken out of ratfn.h since these constructors cause the warning
// "temporary used for non-const struct ratfn & argument"
ratfn::ratfn(const transform_variable z){ *this=polynomial(z); }
ratfn::ratfn(const complex z){ *this=polynomial(z); }
ratfn::ratfn(const double z){ *this=polynomial(z); }

void ratfn::simplify() {
    // cancel factors common to num and den
    num.bless_roots();
    den.bless_roots();

    for( int i=0; i<num.degree(); ++i ) {
	// check each denominator root for closeness to this numerator root
	complex num_root=num.root[i];
	double threshold = 100 * macheps() * ( 1+abs(num_root) );
	for( int j=0; j<den.degree(); ++j ) {
	    if( abs( num_root-den.root[j] ) < threshold ) {
		num.remove_root(i--);
		den.remove_root(j);
		break;
	    }
	}
    }
    valid = polynomial_quotient;
}

int ratfn::read_in(const char *fn) {
    ifstream file(fn);
    if( !file ) {
	cerr <<"Unable to open " << fn << endl;
	return 0;
    }
    file >> *this;
    file.close();
    return 1;
}

void ratfn::pf_valid( int ok ) {
     if (ok)
	 valid |= partial_fraction_sum;
     else {
         valid &= ~partial_fraction_sum;
         delete[n_pf] pf;
     }
}

void ratfn::calculate_partial_fraction_expansion() {
    if( pf_valid() )
	delete pf;
    pf = new partial_fraction[den.degree()];

    {   polynomial tmp=num;
	leading_polynomial = divide( tmp, den );
	}
    n_pf=0; 
    den.bless_roots();
    for( int i=0; i<den.degree(); ++i ) {
	// ! Domain CC does not accept this statement:
	// ! complex root=(pf[n_pf].root=den.root[i]);
	complex root=den.root[i];	pf[n_pf].root=root;
	complex reduced_den = den.reduced_evaluation(i); 
	pf[n_pf].coeff = num(root)/reduced_den;
	int m0 = (pf[n_pf++].multiplicity=den.multiplicity(i));
	if( m0 > 1 ) {	// special case for repeated roots
    	    ratfn H=*this;
	    int i0=i;
	    polynomial reduced_den_poly = den.reduced_polynomial(i);
	    // remove the various powers of the partial fraction.
	    for( int m=m0-1; m>0; m--, ++i ) {
		// calculate the reduced H
		H.num -= pf[n_pf-1].coeff * reduced_den_poly;
		H.num.deflate( root ); // could check that this is small
		H.den.remove_root(i0);
		pf[n_pf].coeff = H.num(root)/reduced_den;
		if( abs(pf[n_pf].coeff) < 1000*macheps() )
		    continue;
		pf[n_pf].root = root;
		pf[n_pf++].multiplicity = m;
	    }  
	}
    }
    // should deallocate unused space in the pf array.
    // (but maybe C++ remembers the amount of space so that
    // "delete[n] pf" is the same as "delete pf") 
    pf_valid(1);
}

polynomial ratfn::reduce() {
    return divide( num, den );
}

ostream& operator<<(ostream& s, const partial_fraction& pf ) {
    return s << pf.coeff << " / ( z -" << pf.root << " ) ^" << pf.multiplicity;
}

ostream& operator<<(ostream& s, const ratfn& h ) {
    return s << h.num << " / " << h.den;
}

void ratfn::print_partial_fraction() {
    bless_partial_fraction();
    for( int i=0; i<n_pf; ++i ) {
	if( i )
	    cout << " + ";
        cout << pf[i];
    }
    cout << endl;
}

istream& operator>>(istream& s, ratfn& h) {
    char c;
    s >> h.num >> c >> h.den;
    if( c != '/' ) {
	cerr << "read a " << c << ", when expecting a /, while trying to read in a ratfn.\n";
	// !! Need an error handler.
	// exit(-1);
    }
    h.valid = polynomial_quotient;
    return s;
}

ratfn operator*( const ratfn &h1, const ratfn &h2 ) {
    ratfn h;
    h.num = h1.num * h2.num;
    h.den = h1.den * h2.den;
    h.simplify();
    return h;
}

polynomial operator+(polynomial& p1, complex& c1) {
// all I need to do is to add the complex to the constant term
    polynomial p;
    p1.bless_coeffs();
    p1.coeff[0] += c1;
    p = p1;
    return(p);
}    

ratfn operator+( const ratfn &h1, const ratfn &h2 ) {
    ratfn h;
    h.num = h1.num * h2.den + h2.num * h1.den;
    h.den = h1.den * h2.den;
    h.simplify();
    return h;
}

ratfn operator-( const ratfn &h1, const ratfn &h2 ) {
    ratfn h;
    h.num = h1.num*h2.den - h2.num * h1.den;
    h.den = h1.den*h2.den;
    h.simplify();
    return h;
}

ratfn operator/( const polynomial &p1, const polynomial &p2 ) {
    ratfn h;
    h.num = p1;
    h.den = p2;
    h.nd_valid(1);
    h.simplify();
    return h;
}

ratfn ratfn::unitCircleMagnitudeSquared(){ // assumes the ratfn is real!!
    // form H=h(z)h(z^-1)/z
    ratfn H = (*this)*(invert_argument(num)/invert_argument(den));
    int extra=num.degree() -den.degree() +1;
    if( extra > 0 )	// add zero roots to the denominator
	H.den.add_zero_roots( extra );
    else  		// add zero roots to the numerator
	H.num.add_zero_roots( -extra );
    return H;
}

complex ratfn::invert_z_transform( int n ) {
    // Assume the Z-transform is causal.
    // (It might have ben easier to separate out z^-1 factors before
    // doing the partial-fraction expansion...)
    bless_partial_fraction();
    if(n<0)
	return 0;
    else if( n==0 )
	return leading_polynomial.coeff[0];

    complex impulse=0;
    for( int i=0; i < n_pf; ++i )
	impulse += pf[i].invert_z_transform(n);
    return impulse;
}

complex ratfn::pf_evaluate( complex z ) {
    bless_partial_fraction();
    complex sum=leading_polynomial(z);
    for( int i=0; i<n_pf; ++i )
	sum += pf[i](z);
    return sum;
}

double ratfn::integrate_mag2_along_unit_circle( double f1, double f2 ) {
    ratfn H=unitCircleMagnitudeSquared();
    double mag2 = H.analytically_integrate_along_C( f1, f2 );
    if( mag2 < 0 )
	mag2 = numerically_integrate_mag2_along_C( f1, f2 );
    return mag2;
}

double ratfn::analytically_integrate_along_C( double f1, double f2 ) {
    // actually calculates just the imaginary part of the integral
    bless_partial_fraction();
    polynomial p = integral(leading_polynomial);
    complex z1=complex( cos(f1), sin(f1) ), z2=complex( cos(f2), sin(f2) );
    double mag2=imag(p(z2))-imag(p(z1));

    for( int i=0; i<n_pf ; ++i ) {
	double term=imag( pf[i].evaluate_integral(z1,z2) );
	if( pf[i].multiplicity == 1 
	  && encircles_more_than_halfway( pf[i].root, z1, z2 ) )
	    term += 2*M_PI*real(pf[i].coeff);
	mag2 += term;
    }
    if( mag2 < 100 * pf_error(f1,f2) )
	mag2 = -1;    // return a negative number if there was numerical trouble
    return mag2;
}

int ratfn::encircles_more_than_halfway( complex z, complex z1, complex z2 ) {
    // see PhD Notebook #2, page 60
    if( norm(z) > 1 )
	return 0;
    complex rot = 1/sqrt(z1*z2);
    if( imag(rot) > 0 ) rot = -rot; 	// I need a specific branch of sqrt()
    if(  real(z*rot) < real( rot*(z1+z2)/2 )  )
	return 0;
    else
	return 1;
}

double ratfn::pf_error( double f1, double f2 ) {
    // estimate the error in the pf representation across the band of interest
    const int npts=10;
    double error=0, deltaf=(f2-f1)/npts, f0=f1+deltaf/2;
    complex z(cos(f0), sin(f0)), dz(cos(deltaf),sin(deltaf));
    for( int i=0; i<npts ; ++i, z*=dz )
	error += abs( (*this)(z) - pf_evaluate(z) );
    return error/npts;
}

double ratfn::numerically_integrate_mag2_along_C( double f1, double f2 ) {
    const int npts=256;
    double mag2=0, deltaf=(f2-f1)/npts, f0=f1+deltaf/2;
    complex z(cos(f0), sin(f0)), dz(cos(deltaf),sin(deltaf));
    for( int i=0; i<npts ; ++i, z*=dz )
	mag2 += norm( (*this)(z) );
    return mag2*deltaf;
}

int ratfn::stable() {
    int all_inside_C = 1;
    den.bless_roots();
    for( int i=0; i<den.degree() && all_inside_C; ++i)
	if( norm(den.root[i]) > 1 )
	    all_inside_C = 0;
    return all_inside_C;
}

double ratfn::h1() {
    if( !stable() )
	 return 1e10;	// !! This should be infinity. !!
    double h1 = abs( invert_z_transform(0) );
    int consecutiveSmallSamples=0;
    for(int n=1; n<10000 && consecutiveSmallSamples<10; ++n){
	double sample = abs( invert_z_transform(n) );
        h1 += sample;
        if( sample < 1e-16 )
            ++consecutiveSmallSamples;
        else
            consecutiveSmallSamples=0;
    }
    return h1;
}
