/*
 * decimate.C
 *  apply a decimation algorithm to a sequence given on the input.
 * usage:
 *	decimate method N
 *    method:	rect = rectangular
 *		tri  = triangular (needs N extra samples)
 *	   XXX  quad = piece-wise quadratic (needs 3N extra samples)
 *	   XXX  cos = raised cosine
 *		Rect = rectangular with multiplication by exp(-j pi/4 t)
 *		Tri = triangular with multiplication by exp(-j pi/4 t)
 *		the name of a file containing a filter transfer function
 *
 *    N is the decimation ratio
 *
 */

#include <stdlib.h>
#include <stream.h>
#include <fstream.h>
#include <complex.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

typedef int (DECIMATOR)(complex);
int N=4;
complex Result=0;
double Wt[]={ 1,  M_SQRT1_2, 0, -M_SQRT1_2, 
	     -1, -M_SQRT1_2, 0,  M_SQRT1_2, 
	      1,  M_SQRT1_2 }; 

int rectangular(complex input) {
    static int n=0;
    static complex acc=0;

    acc += input;

    if( ++n == N ) {
	Result = acc/N;
	n=0;
	acc=0;
	return(1);
    }
    else
	return(0);
}

int Rectangular(complex input) {
    static int n=0, nMod8=0;
    static complex acc=0;

    acc += input * complex(Wt[nMod8], Wt[nMod8+2]);	/* (cos, -sin)  */
    if( ++nMod8==8 )
	nMod8=0;

    if( ++n == N ) {
	Result = acc/N;
	n=0;
	acc=0;
	return(1);
    }
    else
	return(0);
}

int triangular(complex input) {
    static int n=0;
    static int init=1; /* the number of cycles required for initialization */
    static complex acc1=0;
    static complex acc2=0;
    static complex oldacc2=0;

    acc1 += input;
    acc2 += input * ++n;

    if( n == N ){
	Result = (oldacc2 + N*acc1 - acc2)/(N*N);
	oldacc2=acc2;
	acc1=0;
	acc2=0;
	n=0;
	if( init >0 ){
	    --init;
	    return(0);
	}
	else
	    return(1);
    }
    else
	return(0);
}

int Triangular(complex input) {
    static int n=0,nMod8=0;
    static int init=1; /* the number of cycles required for initialization */
    static complex acc1=0;
    static complex acc2=0;
    static complex oldacc2=0;

    complex mod;
    int rc=0;

    mod = input*complex( Wt[nMod8], Wt[nMod8+2] );	/* cos, -sin */
    if( ++nMod8==8 )
	nMod8=0;

    acc1 += mod;
    acc2 += mod * ++n;


    if( n == N ){
	Result = (oldacc2 + N*acc1 - acc2)/(N*N);
	oldacc2=acc2;
	acc1=0;
	acc2=0;
	n=0;
	if( init >0 )
	    --init;
	else
	    rc=1;
    }

    return(rc);
}

int quadratic(complex input) {
    static int n=0;
    input=0;
    return(0);
    }

int cosine(complex input) {
    static int n=0;
    input=0;
    return(0);
    }

/* the variables associated with a filter. Leave them global for now */
int nzeros, npoles;
complex gain;
complex zeros[99], poles[99], state[99];

void init_filter( ifstream &file ){
int i;

file >> nzeros;
file >> npoles;
if( nzeros > npoles ){
    fprintf( stderr, "nzeros cannot exceed npoles.\n" );
    exit( -1 );
    }

file >> gain;
for( i=0; i< nzeros; ++i )
    file >> zeros[i];
for( i=0; i< npoles; ++i )
    file >> poles[i];
file.close();

for( i=0; i< npoles; ++i ){
    state[i] = 0;
    }
}

int filter(complex input) {
    static int n=0;
    int i;
    complex in, out, tmp;

    /* first do the cascade of first-order sections with zeros */
    in = input;
    for( i=0; i<nzeros; ++i ) {
	tmp = in + poles[i]*state[i];
	out = tmp - zeros[i]*state[i];
	state[i]=tmp;
	in=out;
	}
    /* then do the cascade of first-order sections without zeros */
    for( i=nzeros; i<npoles; ++i ) {
	tmp = in + poles[i]*state[i];
	out = state[i];
	state[i]=tmp;
	in=out;
	}

    Result = gain * out;
    if( ++n == N ){
	n=0;
        return(1);
	}
    else
        return(0);
    }

main( int argc, char **argv )
{
    DECIMATOR *decimate;

    --argc;
    if( argc == 0 ){
	fprintf( stderr, "decimate: please give the name of the decimator.\n" );
	exit( -1 );
	}
    if( strncmp( "rectangular", argv[1], strlen(argv[1]) ) == 0 ) 
        decimate = rectangular;
    else if( strncmp( "Rectangular", argv[1], strlen(argv[1]) ) == 0 ) 
        decimate = Rectangular;
    else if( strncmp( "triangular", argv[1], strlen(argv[1]) ) == 0 ) 
        decimate = triangular;
    else if( strncmp( "Triangular", argv[1], strlen(argv[1]) ) == 0 ) 
        decimate = Triangular;
    else if( strncmp( "quadratic", argv[1], strlen(argv[1]) ) == 0 ) 
        decimate = quadratic;
    else if( strncmp( "cosine", argv[1], strlen(argv[1]) ) == 0 ) 
        decimate = cosine;
    else {
	    ifstream file(argv[1]);
	    if( !file ){
		fprintf( stderr, "decimate: %s is not a recognized decimation technique, nor a file containing a filtering function.\n", argv[1] );
		exit( -1 );
		}

	    init_filter( file );
	    decimate = filter;
	    }

    if( --argc>0 ) {
	N = atoi( argv[2] );
	if( argc > 1 )
	    fprintf( stderr, "decimate: extra args ignored.\n" );
    	}

    complex x;
    while( cin >> x )
	if( decimate(x) )
	    cout << Result << endl;
}
