/*
 * Simulate a sigma-delta modulator, with the Schreier structure.
 * usage:
 *    modulator [options] [parameters] fn
 * options are:
 *	l	replace the quantizer with an additive noise source
 *	s	print statistics on the x-signal
 *      u	find and print the stable (constant) input limit 
 *	e	print e
 *	He	print He
 *	x	print x
 *	xy	print x and y
 *	xye	print x, y and e
 *	y	print y (the default unless another output is specified)
 *
 * parameters are:
 *	dc=	the value of the input (normally taken from stdin)
 *      n=	number of quanLevels (2)
 *      N=	number of time steps (1024) [required if -u or dc= specified]  
 *
 * arguments are:
 *	fn	A file containing the description of the modulator
 * 
 */

// !! I would prefer to use NaN, but I can't count on it. This is ugly.
#define NOTDEFINED 4517.107

/* Global variables */
double u=NOTDEFINED, x, y, xLimit=100;
#include <mainhdr.h>
#include "filter.h"
#include "stats.h"


double filtered_error=0, e;
int nLevels=2;
long t=0, tMax=1024;

double sgn( double x ){ 
    return x<0?-1:1;
    }

double mid_tread_quantizer( double x ){ 
    double n = 2*floor(0.5*(x+1));
    if( n > nLevels )
	return nLevels-1;
    else if( n < -nLevels )
	return 1-nLevels;
    else 
	return n;
    }

double mid_riser_quantizer( double x ){ 
    double n = 2*floor(0.5*x)+1;
    if( n > nLevels )
	return nLevels-1;
    else if( n < -nLevels )
	return 1-nLevels;
    else 
	return n;
    }

// I can't find this in libF77: double drandm(int);
// I've used drand48() in past work
double additive_noise( double x ){ 
    return x + rand()/0x40000000-1;
    }

int inputFromStdin(){
    cin >> u;
    return cin.good();
    }

int dcInput(){
    return (t++)<tMax;
    }

void printE(){
    cout << e << endl;
    }

void printHE(){
    cout << y-u  << endl;
    }

void printX(){
    cout << x << endl;
    }

void printY(){
    cout << y << endl;
    }

void printXY(){
    cout << x << '\t' << y << endl;
    }

void printXYE(){
    cout << x << '\t' << y << '\t' << e << endl;
    }

void printNothing(){
    }

main(int argc, char *argv[]){
    double nLevels_d=2,tMax_d=tMax; // !! One day make parseArgs take ints, longs etc.
    int linearModel=0,
        find_x_stats=0,
	findStableInputLimit=0,
	print_e=0,
	print_He=0,
	print_x=0,
	print_xy=0,
	print_xye=0,
	print_y=0;
    // This would be cleaner if C++ allowed array initialization.
    parameterStruct parms[4], *p=parms;
    p->string = "dc";	        p++->parameter = &u;
    p->string = "n";	        p++->parameter = &nLevels_d;
    p->string = "N";		p++->parameter = &tMax_d;
    p->string = (char *)NULL;	p++->parameter = (double *)NULL;
    flagStruct flags[10], *f=flags;
    f->string = "l";      	f++->flag = &linearModel;
    f->string = "s";      	f++->flag = &find_x_stats;
    f->string = "u";      	f++->flag = &findStableInputLimit;
    f->string = "e";      	f++->flag = &print_e;
    f->string = "He";      	f++->flag = &print_He;
    f->string = "x";      	f++->flag = &print_x;
    f->string = "xy";      	f++->flag = &print_xy;
    f->string = "xye";      	f++->flag = &print_xye;
    f->string = "y";      	f++->flag = &print_y;
    f->string = (char *)NULL;	f++->flag = 0;

    parseArgs( argc, argv, parms, flags );
    nLevels = (int)nLevels_d; 
    tMax = (long)tMax_d;

    // Make quantizer point to the correct function.
    double (*quantizer)(double);
    if( linearModel )
	quantizer = additive_noise;
    else { 
	if( nLevels<2 ){
	    ERRORMESG("invalid number of levels, "<< nLevels)
	    exit(-1);
	    }
	else if( nLevels==2 )
	    quantizer = sgn;
	else if( nLevels%2 )	// nLevels is odd
	    quantizer = mid_tread_quantizer;
	else			// nLevels is even
	    quantizer = mid_riser_quantizer;
	}

    // Set up the appropriate input function
    int (*get_input)();
    if( u == NOTDEFINED )
	get_input = inputFromStdin;
    else
	get_input = dcInput;

    // Set up the appropriate output function
    void (*output)();
    if( print_x + print_y + print_e + print_He + print_xy + print_xye > 1 ){
	ERRORMESG("Only one output specification may be given.");
	exit(-1);
	}
    else if( print_xye )
	output = printXYE;
    else if( print_xy )
	output = printXY;
    else if( print_y )
	output = printY;
    else if( print_x )
	output = printX;
    else if( print_e )
	output = printE;
    else if( print_He )
	output = printHE;
    else if( find_x_stats )
	output = printNothing;
    else
	output = printY; // If nothing has been requested, at least print y.

    if( argc == 0 ){
        ERRORMESG("A file containing a modulator description is needed.");
        exit(-1);
        }
    else if( argc > 1 ){
        ERRORMESG("The following arguments followed the filename:");
	while(--argc>0)
	    cerr << *++argv << endl;
	exit(-1);
	}

    ratfn h;
    if( !h.read_in( *argv ) ) exit(-1);
    // cerr << "Error transfer function:\n"<< h << endl;

    transform_variable z;
    filter T(z*(h-1));

    if( !findStableInputLimit ){
	stats st;
	while(get_input()){
	    x = u + filtered_error;
	    y=quantizer(x);
	    e=y-x;
	    output();
	    filtered_error = T(e);
	    if(find_x_stats) st+=x;
	    }

	if(find_x_stats) {
	    if( print_y )
		cerr << st.max_abs << ' ' << sqrt(st.sum2/st.n) << endl;
	    else
		cout << st.max_abs << ' ' << sqrt(st.sum2/st.n) << endl;
	    }
	}
    else {
	/* Starting with zero, increase the input until instability occurs.
	 * Then use a binary search to find the input which results in 
	 * instability. THIS ASSUMES THAT THERE EXISTS ONLY ONE SUCH BREAKPOINT!
	 */
	double deltaInput=0.6231, inputLimit=0;
	int foundaLimit=0;
	for( u=0; deltaInput>.001;  ){
	    T.reset();
	    x=0; filtered_error=0;
	    for(long t=0; t<tMax && fabs(x)<= xLimit; ++t ){
		x = u + filtered_error;
		y=quantizer(x);
		e=y-x;
		filtered_error = T(e);
		}
	    //cout << u <<' '<< tMax <<' '<< fabs(x) << endl;
	    if( t>=tMax ){
		inputLimit = u;
		if( foundaLimit )
		    deltaInput *= 0.5;
		u += deltaInput;
		}
	    else{
		foundaLimit = 1;
		if( u==0 )
		    break;
		deltaInput *= 0.5;
		u -= deltaInput;
		}
	    }
	cout << "u_max= " << inputLimit << endl;
	}
    }
