#include "mainhdr.h"
#include "filter.h"

// Synthesize a high-order noise transfer function with maximum snr
// Usage: (arguments can be in any order)
// optimalNTF order=(order(3)) osr=(osr(64)) w0=(omega_0(0)) timeLimit=10000

// Global variables that make life easier
int order;
double osr = 64;
double w0 = 0;
int optimized=0;
int timeLimit=10000;

/* parameters of the best design so far */ 
double snr=0, u_max, Hinf;
ratfn H;

double sgn( double x ){
    return x<0?-1:1;
    }

void butterworth( cxVector &s ){
    int order = s.size();
    double theta0=M_PI/(2*order), dtheta=2*theta0;
    for( int i=0, i2=0, i2p1=1; i2p1<order; ++i, i2=i2p1+1, i2p1=i2+1 ){
        double theta = theta0 + i*dtheta;
        s[i2] = complex( -sin(theta), cos(theta) );
        s[i2p1] = conj( s[i2] );
        }
    if( order%2 )
        s[i2] = complex(-1,0);
    }

complex zBilinear( complex s, double k=1 ){
    return (k+s)/(k-s);
    }

complex lp2bp( complex s, double b=1 ){
    complex sb=s*(complex)b;
    return -sb+sqrt(sb*sb-1);
    }

void optimizeZeros( double *zeros, int order ){
    double discr, tmp;
    switch(order){
	case 1:
	    *zeros = 0;
	    break;
	case 2:
	    *zeros = 1/sqrt(3.);
	    break;
	case 3:
	    *zeros++ = sqrt(0.6);
	    *zeros = 0;
	    break;
	case 4:
	    discr = sqrt(9./49-3./35);
	    tmp = 3./7;
	    *zeros = sqrt(tmp+discr);
	    *++zeros = sqrt(tmp-discr);
	    break;
	case 5:
	    discr = sqrt(25./81-5./21);
	    tmp = 5./9;
	    *zeros = sqrt(tmp+discr);
	    *++zeros = sqrt(tmp-discr);
	    *++zeros = 0;
	    break;
	case 6: // the following were determined numerically, using MATLAB
	    *zeros = 0.93246935636220;
	    *++zeros = 0.66120925285410;
	    *++zeros = 0.23861957806940;
	    break;
	case 7:
	    *zeros = 0.94910794020414;
	    *++zeros = 0.74153083158296;
	    *++zeros = 0.40584562459724;
	    *++zeros = 0;
	    break;
	case 8:
	    *zeros = 0.96029008224551;
	    *++zeros = 0.79666675114317;
	    *++zeros = 0.52553208300091;
	    *++zeros = 0.18343409158171;
	    break;
	default:
	    cerr << cmdName<<": Sorry, there is no code for optimizing the zeros of "<<order <<"th order modulators\n";
	    exit(-1);
	    break;
	}
    }

ratfn synthesizeNTF( double Hinf ){
    int i;
    cxVector s;
    s.set_size(order);
    ratfn H;
    static double k=0; // use the value from the previous time through as a starting point.

    // Determine the zeros.
    if( w0==0 ){ // Lowpass modulator
	if( !optimized ){ // Coincident zeros
	    for( i=0; i<order; ++i )
		s[i] = 1;
	    }
	else { // Optimized zeros
	    double bandEdge = M_PI/osr;
	    double *optimizedZeros=new double[order], *zeros=optimizedZeros;
	    optimizeZeros(optimizedZeros, order);
	    for( i=0; i<order; ++i ){
		s[i] = expj(bandEdge * *zeros++);
		if(++i<order)
		    s[i] = conj(s[i-1]);
		}
	    delete optimizedZeros;
	    }
	}
    else { // Bandpass modulator
	double pw0 = M_PI*w0;
	if( !optimized ){ // Coincident zeros
	    complex z0 = expj(pw0);
	    for( i=0; i<order; ++i ){
		s[i++] = z0;
		s[i] = conj(z0);;
		}
	    }
	else { // Optimized zeros
	    double bandEdge = M_PI/(2*osr);
	    int order2 = order/2; // !!order better be even!!
	    double *optimizedZeros=new double[order2], *zeros=optimizedZeros;
	    optimizeZeros(optimizedZeros, order2);
	    for( i=0; i<order; ++i ){
		double dw = bandEdge * *zeros++;
		s[i++] = expj(pw0+dw);
		s[i] = conj(s[i-1]);
		if(++i<order){
		    s[i++] = expj(pw0-dw);
		    s[i] = conj(s[i-1]);
		    }
		}
	    delete optimizedZeros;
	    }
	}
    s.set_small_to_zero(1e-12);
    H.num.set_roots(1,s);

    // Iteratively determine the poles to achieve the desired infinity-norm.
    if( w0 != 0 )
	 s.set_size(order/2);
    butterworth(s);
    cxVector z; z.set_size(order);
    if( w0 == 0 ){ // Lowpass design
	if(k==0) k=2*order/(M_PI*(Hinf-1)) + 1;
	double f=1, fprev=1, delta_k=1;
	for( int itnCount=0; 
	  itnCount<100 && fabs(f)>1e-10 && fabs(delta_k)>1e-10;
	  ++itnCount ){
	    for( i=0; i<order; ++i )
		z[i] = zBilinear(s[i],k);
	    z.set_small_to_zero(1e-12);
	    H.den.set_roots(1,z);
	    if(debug)
		cerr << SHOW(k) <<", Hinf="<< H(-1) << endl;
	    f = real(H(-1))-Hinf;
	    if( itnCount==0 )
		delta_k = f/(1+pow(k,-(double)order));
	    else if( fabs( f-fprev ) > 1e-6*(fabs(f)+fabs(fprev)) )
		delta_k = -f*delta_k/(f-fprev);
	    else
		delta_k = -0.5*delta_k; 
	    if( fabs(delta_k) > 0.3 ) // step is out of control
		delta_k = 0.3*sgn(delta_k);
	    
	    k += delta_k;
	    fprev = f;
	    }
        }
    else { // Bandpass design.
	complex *p = new complex[order];
	double b=1/osr;
	double f=1, fprev=1, delta_b=1;
	double z_inf=w0>0.5?1:-1;
	int i2;

	if(k==0) k=1/tan(0.5*M_PI*w0);
	for( int itnCount=0; 
	  itnCount<100 && fabs(f)>1e-10 && fabs(delta_b)>1e-10;
	  ++itnCount ){
	    // LP to BP transformation
	    for( i=0, i2=0; i<order; ++i ){
		p[i++] = lp2bp(s[i2++],b);
		p[i] = 1/p[i-1];
		}
	    // bilinear transformation
	    for( i=0; i<order; ++i )
		z[i] = zBilinear(p[i],k);
	    z.set_small_to_zero(1e-12);
	    H.den.set_roots(1,z);
	    // cerr << "b="<<b <<", Hinf="<< H(z_inf) << '\n';
	    f = real(H(z_inf))-Hinf;
	    if( itnCount==0 )
		delta_b = .1*b;
	    else
		delta_b = -f*delta_b/(f-fprev);
	    
	    b += delta_b;
	    fprev = f;
	    }
	delete p;
        }
    return H;
    }

void simulate( const ratfn &H, long tMax, double &inputLimit, double &mse ){
    transform_variable z;
    filter T(z*(H-1));

    double input, e, x, y;
    /* Starting with zero, increase the input until instability occurs.
     * Then use a binary search to find the input which results in
     * instability. THIS ASSUMES THAT THERE EXISTS ONLY ONE SUCH BREAKPOINT!
     */
    inputLimit=0;
    double deltaInput=0.6231;
    const double xLimit=30;
    int foundaLimit=0;
    long t;
    for( input=0; deltaInput>.001;  ){
	T.reset();
	x=input;
	for(t=0; t<tMax && fabs(x)<= xLimit; ++t ){
	    y=sgn(x);
	    e=y-x;
	    x = input + T(e);
	    }
	//cout << input <<' '<< tMax <<' '<< fabs(x) << endl;
	if( t>=tMax ){
	    inputLimit = input;
	    if( foundaLimit )
		deltaInput *= 0.5;
	    input += deltaInput;
	    }
	else{
	    foundaLimit = 1;
	    if( input==0 )
		break;
	    deltaInput *= 0.5;
	    input -= deltaInput;
	    }
	}
    if( inputLimit == 0 ){ // Don't bother finding the mse
	mse = 9e99;
	return;
	}

    // Simulate the modulator with an input = 90% of the above input limit
    // and determine the mse.
    inputLimit *= 0.9;
    mse=0;
    T.reset();
    x=inputLimit;
    for( t=0; t<tMax; ++t ){
	y=sgn(x);
	e=y-x;
	mse += e*e;
	x = inputLimit + T(e);
	}
    mse = mse / tMax;
    }

class optimizationState {
public:
    int status; // Try using an enum: none, one, two, three, bracketed
    double x0, y0, x1, y1, x2, y2;
    double delta1, delta2;
    optimizationState(){ status=0; }
    int update( double &x, double y );
    };

// Update the optimization state and set x to the next value to try
int optimizationState::update( double &x, double y ){
    int continueOptimization=1;
    if( status == 1 ){ // we only have one point
	if( x > x0 ){
	    x1 = x; y1 = y;
	    }
	else{
	    x1 = x0; y1 = y0;
	    x0 = x; y0 = y;
	    }
	delta1 = x1-x0;
	if( y1 > y0 )
	    x = x1 + delta1;
	else
	    x = 0.5*(x1+x0);
	status = 2;
        return continueOptimization;
	}
    if( status == 2 ){ // we only have two points
	if( x<x0 ){
	    x2 = x1; y2 = y1;
	    x1 = x0; y1 = y0;
	    x0 = x; y0 = y;
	    delta2 = delta1;
	    delta1 = x1-x0;
	    }
	else if( x<x1 ){
	    x2 = x1; y2 = y1;
	    x1 = x; y1 = y;
	    delta1 = x1-x0;
	    delta2 = x2-x1;
	    }
	else{
	    x2 = x; y2 = y;
	    delta2 = x2-x1;
	    }
	if( y1>y0 && y1>y2 ){
	    status = 4;
	    if( delta1 > delta2 )
		x = 0.5*(x0+x1);
	    else
		x = 0.5*(x1+x2);
	    }
	else{
	    status = 3;
	    if( y2 > y0 )
		x = x2 + delta2;
	    else
		x = x0 - delta1;
	    }
        return continueOptimization;
	}

    if( x < x0 ){ // the new data point is to the left 
	x2=x1; y2=y1;
	x1=x0; y1=y0;
	x0=x; y0=y;
	}
    else if( x < x1 ) // the new data point is in the left interval
	if( y>y1 ){ // change the right endpoint
	    x2=x1; y2=y1;
	    x1=x; y1=y;
	    delta2 = x2-x1;
	    }
	else{ // change the left endpoint
	    x0=x; y0=y;
	    delta1 = x1-x0;
	    }
    else if( x < x2 ) // the new data point is in the right interval
	if( y>y1 ){ // change the left endpoint
	    x0=x1; y0=y1;
	    x1=x; y1=y;
	    delta1 = x1-x0;
	    }
	else{ // change the right endpoint
	    x2=x; y2=y;
	    delta2 = x2-x1;
	    }
    else{ // the new data point is to the right
	x0=x1; y0=y1;
	x1=x2; y1=y2;
	x2=x; y2=y;
	}

    if( status==3 && y1>y0 && y1>y2 ) // the peak is now bracketed
	status = 4;

    if( status == 4 )
	// Next try the midpoint of the larger of the intervals.
	// Stop if the interval widths are small.
	if( delta1 > delta2 ){
	    x = 0.5*(x0+x1);
	    if( delta1 < .01 )
		continueOptimization = 0;
	    }
	else{
	    x = 0.5*(x1+x2);
	    if( delta2 < .01 )
		continueOptimization = 0;
	    }
    else
	// The next value to try is in the direction of increasing y
	if( y2 > y0 )
	    x = x2 + delta2;
	else
	    x = x0 - delta1;

    return continueOptimization;
    }

// Return the expected snr. 
// Approximate this by supplying an input = 0.9*u_max and determining
// the mean-square value of e.
double performance( double newHinf ){
    double w1, w2;
    if( w0 == 0 ){ w1 = 0; w2 = M_PI/osr; }
    else{ w1 = M_PI*(w0-0.5/osr); w2 = M_PI*(w0+0.5/osr); }
    ratfn newH = synthesizeNTF( newHinf );
    double newUmax , mse;
    simulate( newH, timeLimit, newUmax, mse );
    double newSnr = M_PI*newUmax*newUmax /
     ( mse * newH.integrate_mag2_along_unit_circle(w1,w2) );
    if(debug)
    cerr << SHOW(newHinf) << SHOW(newUmax) << SHOW(mse) << SHOW(newSnr) << endl;
    if( newSnr > snr ) { // Save this design
	    H = newH;
	    snr = newSnr;
	    u_max = newUmax;
	    Hinf = newHinf;
	    }
    return newSnr;
    }

main(int argc, char *argv[]){
    double order_d=3;
    double timeLimit_d=timeLimit;

    // Parse the arguments.
    struct parameterStruct parameters[5], *p=parameters;
    p->string = "order";	p++->parameter = &order_d;
    p->string = "osr";		p++->parameter = &osr;
    p->string = "w0";		p++->parameter = &w0;
    p->string = "timeLimit";	p++->parameter = &timeLimit_d;
    p->string = (char*)NULL;	p++->parameter = (double *)NULL;
    // Flags:
    struct flagStruct flags[2], *f=flags;
    f->string = "opt";		f++->flag = &optimized;
    f->string = "debug";	f++->flag = &debug;
    f->string = (char*)NULL;	f++->flag = 0;
    parseArgs( argc, argv, parameters, flags );
    order = (int)order_d;
    timeLimit = (int)timeLimit_d;
    // cerr << "order="<<order<<", osr="<<osr<<", Hinf="<<Hinf<<", w0="<<w0<<'\n';
    if( argc!=0 ){
	cerr << "The following arguments were not recognized:\n";
	while( argc-->0 )
	    cerr << *argv++ << "\n";
	exit( -1 );
	}
    if( w0!=0 && order%2!=0 ){
	cerr << "n must be even for a bandpass modulator\n";
	exit( -1 );
	}

    // Adjust Hinf for maximum SNR
    optimizationState state;
    state.x0 = 1; state.y0 = 0; //  
    state.status = 1;
    double newHinf = 1.5;
    while( state.update(newHinf, performance(newHinf)) )
	;

    // Print out the results.
    cout.precision(10);
    cout << H.num << "\n/\n" << H.den << "\n\n";
    cout << "w0= "<< w0 << endl;
    cout << "osr= "<< osr << endl;
    cout << '\n';
    cout << "H_inf= " << Hinf << endl;
    cout << "h_1= " << H.h1() << endl;
    cout << "H_2= " << 
      sqrt(H.integrate_mag2_along_unit_circle(0,M_PI) / M_PI) << endl;
    cout << "u_max= " << u_max << endl;
    cout << "snr_expected= " << 10*log10(snr) << " dB\n";
    }
