/*
 * partition [-{sdvr}] filter-file n
 *   s	suppress statistics
 *   d	suppress detail: print only valid output patterns
 *   v	don't print vaild output patterns
 *   r	restrict the input to the range [-0.7,0.7]
 *
 * prints out the partitioning of [-1,+1] for n time samples
 * for a sigma-delta modulator with my standard structure.
 *
 * Author: Richard Schreier
 * For an explanation of the basic technique, see pg 2 of PhD notebook #2
 */

#include <libc.h>
#include <stream.h>
#include "filter.h"

#define MAX_LENGTH 1024

int n;		// the number of time samples
ratfn h;	// the error nominal transfer function
char pattern[MAX_LENGTH];	// a string holding the current output pattern
double a, *b;			// terms used in the inequalities
// statistics gathering
char output_fmt[] = "%s %18.15f %18.15f %18.15f\n";
int interval_count=0;
#define DETAIL 1
#define VALID 2
#define STATS 4
int output_control=DETAIL+VALID+STATS;
double max_width=0, min_width=2, width2_sum=0, width3_sum=0;
char *prog_name;

struct range { double min, max; } restrict={-1.,+1.};

main( int argc, char *argv[] ){
    // parse the arguments
    prog_name = argv[0];
    if( argv[1][0] == '-' ) {
	char c;
	while( c= *++(argv[1]) )
	    switch( c ){
		case 's':
		    output_control&=~STATS;
		    break;
		case 'd':
		    output_control&=~DETAIL;
		    break;
		case 'v':
		    output_control&=~VALID;
		    break;
		case 'r':
		    restrict.min=-0.7;
		    restrict.max= 0.7;
		    break; 
		default: cerr << prog_name <<": unknown - argument:"<< c <<endl;
		    exit( -1 );
		}
	--argc; ++argv;
	}
    if( argc != 3 ){
	cerr << prog_name << ": Please supply a file name and the 'n' argument\n";
	exit( -1 );
	}
    if( !h.read_in(argv[1]) ) exit(-1);
    n = atoi(argv[2]);	// !! could use strstream?
    if( n >= MAX_LENGTH ) {
	cerr <<prog_name<<": the duration is too large (>="<<MAX_LENGTH<<").\n";
	exit( -1 );
	}

    void explore_patterns();
    explore_patterns();


    void print_stats();
    if( output_control&STATS ) print_stats();

    } // main()

double min( double a, double b ){
    return( a<b?a:b );
    }

// Determine the range of u which yields each possible y-pattern
void explore_patterns(){
    transform_variable z;
    cxfilter T(z*(h-1));

    // b[] is n values of the step response of 1/H.
    // Find it by simulating the z*(H-1) loop.
    // b[0] is the input weighting for the first time instant.
    b = new double[n];
    {
	double x=0;
	for( int i=0; i<n; ++i ){
	    x += 1;
	    b[i] = x;
	    x = real(T.simulate(-x));
	    }
	}

    // Initialization
    struct range  u;
    u.min = restrict.min, u.max=restrict.max;
    pattern[n] = '\0';

    // begin a binary tree search for the range associated with each pattern
    void apply_input( cxfilter&, double, int, range );
    T.reset(); a=0;
    apply_input( T, -1, 0, u );
    T.reset(); a=0;
    apply_input( T, +1, 0, u );

    // Cleanup
    delete [n] b;
    } // explore_patterns()

void apply_input( cxfilter &T, double y, int depth, range u ){
    if(y==1)
	pattern[depth]='+';
    else
	pattern[depth]='-';

    // compute new range
    range u1=u;
    // consistency check: sgn(a+bu) = y  => an inequality in u
    if( y*b[depth]>0 )	// determines whether inequ. is u > (-a/b) or u < (-a/b)
	u1.min = max( u.min, -a/b[depth]);
    else
	u1.max = min( u.max, -a/b[depth]);
    ++depth;

    // output the range, or explore further
    if( u1.min >= u1.max ){		// prune the tree (null branch)
	if(output_control&DETAIL){
	    for( int i=depth; i<n; ++i ) 
		pattern[i] ='*';
	    printf( output_fmt, pattern, min(u1.min,u.max), max(u1.max,u.min), 0.0 );
	    }
	}
    else if( depth==n ){		// at the bottom of the tree
	double width = u1.max-u1.min;
	void gather_stats(double);
	gather_stats( width );
	if(output_control&VALID)
	    printf( output_fmt, pattern, u1.min, u1.max, width );
	}
    else{				// explore further
	// a is the constant term resulting from the y-pattern
	a = real(T.simulate(y-a));
	double saved_a=a;
	struct state saved_state=T.get_state();
	apply_input( T, -1, depth, u1 );
	a = saved_a;
	T.set_state(saved_state);
	apply_input( T, +1, depth, u1 );
	}
    } // apply_input()

void gather_stats(double width){
    ++interval_count;
    double width2 = width*width;
    width2_sum += width2;
    width3_sum += width2*width;
    if( width > max_width )
	max_width=width;
    if( width < min_width )
	min_width=width;
    }

void print_stats(){
    double range_width = restrict.max-restrict.min;
    printf( "Number of intervals: %d\n", interval_count );
    printf( "Maximum interval width: %g\n", max_width );
    double avg_width = range_width / interval_count;
    printf( "Average interval width: %g\n", avg_width );
    printf( "Minimum interval width: %g\n", min_width );
    double std_dev = sqrt(width2_sum/interval_count - avg_width*avg_width);
    printf( "Standard deviation of interval width: %g\n", std_dev );
    double mse = width3_sum / (12*range_width);
    printf( "Mean squared error for uniform distribution over [%g,%g]: %g\n", restrict.min, restrict.max, mse );
    }
