# A csh script for running a series of SNR simulations on a given modulator.
# Usage:
# snrCheck modulator -{fp} [parameters]
set noglob

# The default settings:
set p=-20	# power of input tone, in dB
set f=3		# f cycles in
set t=2048	# t time units 
set R=64	# oversampling ratio of R, R <= t/(2f), R must divide t.
set w0=0	# normalized center frequency, 1 -> fs/2
set n=2		# n quantizer levels
set vary_p	# vary input power
set vary_f	# vary frequency
set dec=none	# no decimation
set k=1		# factor for increasing the length of the simulation

if ( $#argv == 0 ) then
    echo "Usage:"
    echo "$0 modulator-name {-fp} {parameters}"
    echo " where the parameters and their defaults are:"
    echo "  R=oversampling ratio($R)"
    echo "  w0=normalized center frequency($w0)"
    echo "  n=quantizer levels($n)"
    echo "  p=input power($p)"
    echo "  f=frequency offset, in bins($f)"
    echo "  t=duration($t)"
    echo "  k=scaling factor($k) for f and t"
    echo "  dec=decimation($dec)"
    echo "The frequency is the number of complete cycles in the course of the simulation."
    echo "If the modulator file is executable, it is run, otherwise it is assumed to contain the ntf."
    exit
endif

# Handle the arguments
set mod=$argv[1]
if( ! -r $mod ) then
    echo The modulator file $mod is not readable.
    exit
endif
set rfile = $mod.r
set pfile = $mod.p
set ffile = $mod.f

shift argv
foreach ARG ($argv)
    switch ($ARG)
    case "-p":
	unset vary_p
	breaksw
    case "-f":
	unset vary_f
	breaksw
    case n=*:
	set n=`substr $ARG 3`
	breaksw
    case p=*:
	set p=`substr $ARG 3`
	breaksw
    case f=*:
	set f=`substr $ARG 3`
	breaksw
    case t=*:
        set t=`substr $ARG 3`
	breaksw
    case k=*
	set k=`substr $ARG 3`
	breaksw
    case R=*
	set R=`substr $ARG 3`
	breaksw
    case w0=*
	set w0=`substr $ARG 4`
	breaksw
    case dec=*
	set dec=`substr $ARG 5`        # I haven't figured out how to put this in
	echo "decimation not supported, yet."
	exit 1
	breaksw
    default:
	echo $0": unknown argument" $ARG
	exit 1
	breaksw
    endsw
end
if( ! -x $mod ) then
    set mod = "modulator $mod n=$n"
endif

@ t = $t * $k
@ f = $f * $k

if( $w0 == 0 ) then
    @ f1 = 0
    @ f2 = $t / ( 2 * $R )
else
    set f0  = `matheval $w0 $t * 2 / p`
    @ f += $f0
    @ f1 = $f0 - $t / ( 4 * $R )
    @ f2 = $f0 + $t / ( 4 * $R )
endif
if( $f > $f2 ) then
    echo "Semantic error. With an oversampling ratio of $R,"
    echo "the frequency must be no more than $f2."
    exit 1
endif


echo "$0 $mod" >! $rfile
echo "The number of quantizer levels is $n." >> $rfile
echo "The input is a ${p}dB tone, with $f cycles in $t samples." >> $rfile
echo "The oversampling ratio is $R and the center frequency is $w0." >> $rfile

if( $?vary_p ) then
    echo "Doing SNR vs input power simulations with $mod." >> $rfile
    rm -f $pfile
    foreach inPwr ( -2 -4 -6 -8 -10 -20 -30 -40 -50 -60 -70 )
        set snr = `sine N=$t f=$f rms=${inPwr}dB | $mod | spectralAnalysis -w f0=$f f1=${f1} f2=${f2} N=$t`
	echo $inPwr $snr >>! $pfile
    end
endif

if( $?vary_f ) then
    echo "Doing SNR vs frequency simulations for modulator $mod." >> $rfile
    rm -f $ffile
    foreach f ( 0 3 6 9 12 15 )
	@ f = $f * $k
        set snr = `sine N=$t f=$f rms=${p}dB | $mod | spectralAnalysis -w f0=$f f1=0 f2=${f2} N=$t`
	echo $f $snr >>! $ffile
    end
endif

echo "Done simulation." >> $rfile
