/* SpectralAnalysis
 *  Calculate the signal-to-noise ratio assuming a tone input.
 * usage:
 *	spectralAnalysis [-wg] w0=toneBin f1=lowBin f2=highBin N=fftLength skip= numberToSkip(0)
 *
 */

#include <complex.h>
#include "mainhdr.h"
#define binNo(f) f<0?f+Len:f

// Global variables
int window=0;
int f0, f1, f2, Len, t1;
complex *x;

int semantics(){
    int errors=0;
    if( f0==-1 ){
        ERRORMESG("Please supply a f0= argument");
        ++errors;
        }
    if( f1==-1 ){
        ERRORMESG("Please supply a f1= argument");
        ++errors;
        }
    if( f2==-1 ){
        ERRORMESG("Please supply a f2= argument");
        ++errors;
        }
    if( Len==-1 ){
        ERRORMESG("Please supply a N= argument");
        ++errors;
        }
    if(errors) return errors;

    if( f1 > f2 ){
        int tmp=f2;
        f2=f1;
        f1=tmp;
        }
    if( f0<f1 || f0>f2 ){
        ERRORMESG("Semantic error, f0("<< f0 <<") not between f1("<< f1
         << ") and f2("<< f2 <<").\n");
        ++errors;
        }
    if( abs(f0)>Len/2 ){
        ERRORMESG("Semantic error, frequency (f0="<< f0 <<") out of range ("
         << Len/2 <<")\n");
        ++errors;
        }
    if( abs(f1)>Len/2 ){
        ERRORMESG("Semantic error, frequency (f1="<< f1 <<") out of range ("
         << Len/2 <<")\n");
        ++errors;
        }
    if( abs(f2)>Len/2 ){
        ERRORMESG("Semantic error, frequency (f2="<< f2 <<") out of range ("
         << Len/2 <<")\n");
        ++errors;
        }
    if( f1 == -Len/2 )
        ++f1;   /* to ensure fs/2 isn't counted twice */
    return errors;
    }

main( int argc, char *argv[] ) {
    int get_cx_vector();
    void fft(), snr(), doHanningWindow();
    double f0_d=-1, f1_d=-1, f2_d=-1, Len_d=-1, t1_d=0;
    // Parse the arguments.
    struct parameterStruct parameters[6], *p=parameters;
    p->string = "f0";		p++->parameter = &f0_d;
    p->string = "f1";		p++->parameter = &f1_d;
    p->string = "f2";		p++->parameter = &f2_d;
    p->string = "N";		p++->parameter = &Len_d;
    p->string = "skip";		p++->parameter = &t1_d;
    p->string = (char*)NULL;	p++->parameter = (double *)NULL;
    // Flags:
    struct flagStruct flags[3], *f=flags;
    f->string = "g";        	f++->flag = &debug;
    f->string = "w";        	f++->flag = &window;
    f->string = (char*)NULL;    f++->flag = 0;
    parseArgs( argc, argv, parameters, flags );
    f0 = (int)f0_d;
    f1 = (int)f1_d;
    f2 = (int)f2_d;
    Len = (int)Len_d;
    t1 = (int)t1_d;
    if(semantics()) return -1;

    if( get_cx_vector() < 0 ) {
    	perror( cmdName );
    	exit( -1 );
    	}
    fft();
    if( window ) 
	doHanningWindow();
    snr();
    return 0;
    }

/* read data and allocate a cx vector */
int get_cx_vector() {
    complex input;

    x = new complex[Len];
    if( x == NULL ) {
	ERRORMESG("Unable to allocate storage for "<< Len <<" complex numbers.\n");
	return -1;
	}

    int count=-t1;
    while( count<0 && cin >> input )
	count++;
    if( count == 0 ) 
	while( count<Len && cin >> input )
	    x[count++] = input;

    if( count < Len )
	ERRORMESG("Appended "<< Len-count << " zeroes to the input\n");
    if( debug ) /* echo the input */
        for( int i=0; i<Len; ++i )
	    cout << x[i] << "\n";
    return 0;
}

/* Do an in-place decimation-in-time fft */
/* See Figure 6.10 of "Digital Signal Processing" by Oppenheim and Schafer. */
void fft() {
    void reorder( complex*, int, int );
    void butterfly( complex&, complex& );

    register int i, j;
    int log2n=0;
    register int increment, offset;
    double theta;
    complex w, w0;

    for( int tmp=Len; (tmp >>= 1) != 0 ; log2n++ ); /* n = 2^log2n */
    reorder( x, Len, log2n );

    /* do the fft */
    increment = 2; offset = 1; 
    while( log2n-- > 0 ) { 
    	/* this is the first iteration of every stage, 
    	   it has been separated out since w is one for the first itn */
    	for( i=0; i<Len; i+=increment )
    		butterfly( x[i], x[i+offset] );

    	theta = M_PI / offset;
    	w0 = complex( cos(theta), -sin(theta) );
    	w = w0;
    	for( j=1; j<offset; ++j ) {
    		for( i = j + offset; i<Len; i+=increment ) {
    			x[i] *= w;
    			butterfly( x[i-offset], x[i] );
    			}
    		w *= w0;
    		}

    	offset = increment;
    	increment *= 2;
    	}
}

/* order the input vector by the bit-reversed indices */
void reorder( complex* x, int n, int m ) { 
    void cx_swap( complex&, complex& );
    int bit_reverse( int, int );

    for( int i=1; i<n; ++i ) {
    	int j = bit_reverse( i, m );
    	if( i<j )
    		cx_swap( x[i], x[j] );
    	}
}

/* there may be a better way, since I know that the i's are sequential...*/
int bit_reverse( register int i, register int m ) {
    register int ir;
    ir = i&1;
    while( --m > 0 ) {
    	i = i>>1;
    	ir = ir<<1 | i&1;
    	}
    return( ir );
}

/* exchange two complex numbers */
void cx_swap( complex& x, complex& y ) {
    complex tmp;
    tmp = x;
    x = y;
    y = tmp;
}

/* do the basic butterfly operation */
void butterfly( complex &x, complex &y ) {
    complex xtmp;
    xtmp = x;
    x = xtmp + y;
    y = xtmp - y;
}

/* do a Hanning window, by x'(k) = -x(k-1) + 2*x(k) - x(k+1) */
/* Note this changes the scaling of the fft */
/* Since I don't want to smear the tone frequency, I have to do
    some funny stuff with the bins on either side.... */
void doHanningWindow(){ 
    void unwin(int);
    complex x0=x[0], xOld=x[Len-1];

    for( int i=0; i<Len-1; ++i ) {
	complex xtmp = x[i];
	/* this won't compile!!! x[i] = 2*xtmp - xOld - x[i+1]; */
	x[i] = -xOld + 2*xtmp - x[i+1];
	xOld = xtmp;
	}
    x[i] = -xOld + 2*x[i] - x0;

    unwin(f0); /* ?perhaps I shouldn't smear the tone in the first place? */
    }

void unwin(int f){ // !! THIS IS INCORRECT FOR f=1 SINCE THE f=0 bin overlaps
    int fm1=binNo(f-1),fp1=binNo(f+1);
    f=binNo(f);
    complex z = (-x[fm1] + x[f] - x[fp1])/4;
    x[fm1] += z;
    x[f] = z*sqrt(6);
    x[fp1] += z;
    }

void snr(){
    double noisePwr=0, signalPwr;

    /* translate negative frequencies into fft bin indices */
    signalPwr = norm( x[binNo(f0)] );

    for( int f=f1; f<=f2; ++f ){	
	double x2=norm( x[binNo(f)] );
	if( f==f0 )
	    continue;
	if( f==0 || f==Len/2 )
	    /* DC and fs/2 count half as much as other frequencies */
	    x2 /= 4.0;
	noisePwr += x2;
	}

    cout.precision(3);
    if( noisePwr!=0 )
	cout << 10 * log10(signalPwr/noisePwr) << endl;
    else if( signalPwr!=0 )
	cout << MAXDOUBLE << endl;
    else
	cout << "0.0\n";
    }
