#include <libc.h>
#include <fstream.h>
#include "ratfn.h"
#include "parseArgs.h"

// synthesize a high-order noise transfer function with a prescribed infinity-norm.
// Usage: (arguments can be in any order)
// synthesizeNTF n=(order(3)) R=(OSR(64)) Hinf=(desired infinity-norm(1.5)) w0=(omega_0(0))

void butterworth( cxVector &s ){
    int order = s.size();
    double theta0=M_PI/(2*order), dtheta=2*theta0;
    for( int i=0, i2=0, i2p1=1; i2p1<order; ++i, i2=i2p1+1, i2p1=i2+1 ){
        double theta = theta0 + i*dtheta;
        s[i2] = complex( -sin(theta), cos(theta) );
        s[i2p1] = conj( s[i2] );
        }
    if( order%2 )
        s[i2] = complex(-1,0);
    }

complex zBilinear( complex s, double k=1 ){
    return (k+s)/(k-s);
    }

complex lp2bp( complex s, double b=1 ){
    complex sb=s*(complex)b;
    return -sb+sqrt(sb*sb-1);
    }

char *cmdName;

void optimizeZeros( double *zeros, int order, int flag ){
    double discr, tmp;
    switch(order){
	case 1:
	    *zeros = 0;
	    break;
	case 2:
	    *zeros = 1/sqrt(3.);
	    break;
	case 3:
	    *zeros++ = sqrt(0.6);
	    *zeros = 0;
	    break;
	case 4:
	    if(flag){ // Put two zeros in the center
		*zeros = 0;
		*++zeros = sqrt(9./11);
	        }
	    else {
		discr = sqrt(9./49-3./35);
		tmp = 3./7;
		*zeros = sqrt(tmp+discr);
		*++zeros = sqrt(tmp-discr);
		}
	    break;
	case 5:
	    discr = sqrt(25./81-5./21);
	    tmp = 5./9;
	    *zeros = sqrt(tmp+discr);
	    *++zeros = sqrt(tmp-discr);
	    *++zeros = 0;
	    break;
	case 6: // the following were determined numerically, using MATLAB
	    *zeros = 0.93246935636220;
	    *++zeros = 0.66120925285410;
	    *++zeros = 0.23861957806940;
	    break;
	case 7:
	    *zeros = 0.94910794020414;
	    *++zeros = 0.74153083158296;
	    *++zeros = 0.40584562459724;
	    *++zeros = 0;
	    break;
	case 8:
	    *zeros = 0.96029008224551;
	    *++zeros = 0.79666675114317;
	    *++zeros = 0.52553208300091;
	    *++zeros = 0.18343409158171;
	    break;
	default:
	    cerr << cmdName<<": Sorry, there is no code for optimizing the zeros of "<<order <<"th order modulators\n";
	    exit(-1);
	    break;
	}
    }

main(int argc, char *argv[]){
    int order; double order_d=3;
    double R = 64;
    double Hinf = 1.5;
    double w0 = 0;

    int optimized=0, optimized2=0;

    // Parse the arguments.
    struct parameterStruct parameters[5], *p=parameters;
    p->string = "n";		p++->parameter = &order_d;
    p->string = "R";		p++->parameter = &R;
    p->string = "Hinf";		p++->parameter = &Hinf;
    p->string = "w0";		p++->parameter = &w0;
    p->string = (char*)NULL;	p++->parameter = (double *)NULL;
    // Flags:
    struct flagStruct flags[3], *f=flags;
    f->string = "opt2";		f++->flag = &optimized2;
    f->string = "opt";		f++->flag = &optimized;
    f->string = (char*)NULL;	f++->flag = 0;
    parseArgs( argc, argv, parameters, flags );
    order = (int)order_d;
    if( optimized2 ) optimized = 1;
    // cerr << "order="<<order<<", R="<<R<<", Hinf="<<Hinf<<", w0="<<w0<<'\n';
    if( argc!=0 ){
	cerr << "The following arguments were not recognized:\n";
	while( argc-->0 )
	    cerr << *argv++ << "\n";
	exit( -1 );
	}
    if( w0!=0 && order%2!=0 ){
	cerr << "n must be even for a bandpass modulator\n";
	exit( -1 );
	}


    ratfn H;
    int i;
    cxVector s;
    s.set_size(order);

    // Determine the zeros.
    if( w0==0 ){ // Lowpass modulator
	if( !optimized ){ // Coincident zeros
	    for( i=0; i<order; ++i )
		s[i] = 1;
	    }
	else { // Optimized zeros
	    double bandEdge = M_PI/R;
	    double *optimizedZeros=new double[order], *zeros=optimizedZeros;
	    optimizeZeros(optimizedZeros, order, optimized2);
	    for( i=0; i<order; ++i ){
		s[i] = expj(bandEdge * *zeros++);
		if(++i<order)
		    s[i] = conj(s[i-1]);
		}
	    delete optimizedZeros;
	    }
	}
    else { // Bandpass modulator
	double pw0 = M_PI*w0;
	if( !optimized ){ // Coincident zeros
	    complex z0 = expj(pw0);
	    for( i=0; i<order; ++i ){
		s[i++] = z0;
		s[i] = conj(z0);;
		}
	    }
	else { // Optimized zeros
	    double bandEdge = M_PI/(2*R);
	    int order2 = order/2; // !!order better be even!!
	    double *optimizedZeros=new double[order2], *zeros=optimizedZeros;
	    optimizeZeros(optimizedZeros, order2, optimized2);
	    for( i=0; i<order; ++i ){
		double dw = bandEdge * *zeros++;
		s[i++] = expj(pw0+dw);
		s[i] = conj(s[i-1]);
		if(++i<order){
		    s[i++] = expj(pw0-dw);
		    s[i] = conj(s[i-1]);
		    }
		}
	    delete optimizedZeros;
	    }
	}
    s.set_small_to_zero(1e-12);
    H.num.set_roots(1,s);

    // Iteratively determine the poles to achieve the desired infinity-norm.
    if( w0 != 0 )
	 s.set_size(order/2);
    butterworth(s);
    cxVector z; z.set_size(order);
    if( w0 == 0 ){ // Lowpass design
	double k=2*order/(M_PI*(Hinf-1)) + 1;
	double f=1, fprev=1, delta_k=1;
	for( int itnCount=0; 
	  itnCount<100 && fabs(f)>1e-10 && fabs(delta_k)>1e-10;
	  ++itnCount ){
	    for( i=0; i<order; ++i ) {
		z[i] = zBilinear(s[i],k);
            }
	    z.set_small_to_zero(1e-12);
	    H.den.set_roots(1,z);
	    // cerr << "k="<<k <<", Hinf="<< H(-1) << '\n';
	    f = real(H(-1))-Hinf;
	    if( itnCount==0 )
		delta_k = f/(1+pow(k,-(double)order));
	    else
		delta_k = -f*delta_k/(f-fprev);
	    
	    k += delta_k;
	    fprev = f;
	    }
        }
    else { // Bandpass design.
	complex *p = new complex[order];
	double k=1/tan(0.5*M_PI*w0), b=1/R;
	double f=1, fprev=1, delta_b=1;
	double z_inf=w0>0.5?1:-1;
	int i2;

	for( int itnCount=0; 
	  itnCount<100 && fabs(f)>1e-10 && fabs(delta_b)>1e-10;
	  ++itnCount ){
	    // LP to BP transformation
	    for( i=0, i2=0; i<order; ++i ){
		p[i++] = lp2bp(s[i2++],b);
		p[i] = 1/p[i-1];
		}
	    // bilinear transformation
	    for( i=0; i<order; ++i ) {
		z[i] = zBilinear(p[i],k);
            } 
	    z.set_small_to_zero(1e-12);
	    H.den.set_roots(1,z);
	    // cerr << "b="<<b <<", Hinf="<< H(z_inf) << '\n';
	    f = real(H(z_inf))-Hinf;
	    if( itnCount==0 )
		delta_b = .1*b;
	    else
		delta_b = -f*delta_b/(f-fprev);
	    
	    b += delta_b;
	    fprev = f;
	    }
	delete p;
        }

    // Print out the results.
    H.num.sort_roots();
    H.den.sort_roots();
    cout.precision(10);
    cout << H.num << "\n/\n" << H.den << "\n\n";
    cout << "w0= "<< w0 << '\n';
    cout << "R= "<< R << '\n';
    cout << '\n';
    cout << "N0^2= ";
    if( w0 == 0 )
	cout << 10*log10(H.integrate_mag2_along_unit_circle(0,M_PI/R)/(3*M_PI)) << " dB\n";
    else
	cout << 10*log10(H.integrate_mag2_along_unit_circle((w0-0.5/R)*M_PI,(w0+0.5/R)*M_PI)/(3*M_PI)) 
	  << " dB\n";
    cout << "H_inf= " << Hinf << "\n";
    cout << "h_1= " << H.h1() << "\n";
    cout << "H_2= " << 
      sqrt(H.integrate_mag2_along_unit_circle(0,M_PI) / M_PI) << ".\n";
    }
