/* fft [-{Bb}{mMrz}{i}{u}{w}{G}] [file]
 * performs an discrete fourier transform of the data sequence from the 
 * indicated file (stdin default).  The input may be real or complex,
 * the latter are written as (real,imag).
 *
 * Options:
 *  b - input and output are binary, input consists of double
 *  B - input and output are binary, input consists of complex
 *
 *  m - output is the magnitude of each fft bin
 *  M - output is the magnitude squared of each fft bin
 *  r - output is the real part of each fft bin
 *  z - output is the complex value of each fft bin
 *
 *  c - centre the output spectrum [-pi,pi] instead of [0,2*pi)
 *
 *  i - do the inverse transform
 *
 *  u - do a unitary version of the transform ( mpy by 1/sqrt(N) )
 *
 *  w - apply a Hanning window to the input
 *
 *  G - debugging. Echo the input
 *
 */

#include <stdlib.h>
#include <stream.h>
#include <complex.h>
#include <sys/file.h>
extern int open( char*, int );

enum boolean { false, true }; /* false=0, true=1 */


/****************************************/
/*           Global Variables           */
/****************************************/
complex *x;
int n; 
enum boolean debug=false, center=false;
void (*output_it)( const complex & );
int (*get_input)(complex &);

main( int argc, char **argv )
{
    enum boolean unitary=false, window=false, inverse=false;
    enum boolean bits_out=false;
    enum {DOUBLE, COMPLEX, MAG, MAG_SQUARED} output_type=MAG;
    void print_mag( const complex& ), print_mag2( const complex& ), print_cx( const complex& ), print_double( const complex& );
    void write_mag( const complex& ), write_mag2( const complex& ), write_cx( const complex& ), write_double( const complex& );
    int ascii_in( complex& ), fbits_in( complex& ), zbits_in( complex& );
    int get_cx_vector();
    void fft();

    get_input = ascii_in;
    /* process the arguments */
    while (--argc > 0 && (++argv)[0][0]=='-') 
    	{
    	char *cp = argv[0] + 1;
    	while (*cp) switch (*cp++)
	    {
	    case 'm':
		output_type = MAG;
		break;
	    case 'M':
		output_type = MAG_SQUARED;
		break;
	    case 'r':
		output_type = DOUBLE;
		break;
	    case 'z':
		output_type = COMPLEX;
		break;
	    case 'i':
		inverse = true;
		/* this is a potential source of troble */
		output_type = DOUBLE;
		break;
	    case 'u':
		unitary = true;
		break;
	    case 'b':
		get_input = fbits_in;
		bits_out = true;
		break;
	    case 'B':
		get_input = zbits_in;
		bits_out = true;
		break;
	    case 'w':
		window = true;
		break;
	    case 'c':
		center = true;
		break;
	    case 'G':
		debug = true;
		break;
	    default:
		fprintf( stderr, "unknown switch %c\n", *(cp-1) );
	    }
    }

/* I'll not worry about file input anymore.
    if( argc > 0 ) { // take input from a file
	int fd;
    	if( (fd = open( argv[0], O_RDONLY )) < 0 ) {
    		perror( argv[0] );
    		exit( -1 );
    		}
	cin = *new istream( fd );
    	}
*/

    /* Decide on the output function */
    switch( output_type ) {
	case MAG:
	    output_it = bits_out?write_mag:print_mag;
	    break;
	case MAG_SQUARED:
	    output_it = bits_out?write_mag2:print_mag2;
	    break;
	case DOUBLE:
	    output_it = bits_out?write_double:print_double;
	    break;
	case COMPLEX:
	    output_it = bits_out?write_cx:print_cx;
	    break;
	default:
	    fprintf( stderr, "fft: internal error. 'output_type' unknown.\n" );
	    exit( -1 );
	}

    /* get data */
    if( get_cx_vector() < 0 ) {
    	perror( "fft:" );
    	exit( -1 );
    	}

    /* do the fft */
    fft();

    if( unitary || inverse || window ) { /* do some scaling */
    	double scaling;
	if( unitary )
    	    scaling = 1.0 / sqrt( n );
	else if( inverse )
	    scaling =  1.0 / n;
	else
	    scaling =  1.0;
	
	if( window )
	    scaling *= 0.25;

    	for( int i=0; i<n; ++i )
    		x[i] *= scaling;
    	}

    if( inverse ) {
        void cx_swap( complex&, complex& );
        for( int i=1; i<n/2; ++i )
            cx_swap( x[i], x[n-i] );
	}

    if( window ) { 
      /* do a Hanning window, by x'(k) = 0.25 * ( -x(k-1) + 2*x(k) - x(k+1) ) */
	complex x0 = x[0], xOld, xtmp;
	xOld = x[n-1];
        for( int i=0; i<n-1; ++i ) {
            xtmp = x[i];
	    /* this won't compile!!! x[i] = 2*xtmp - xOld - x[i+1]; */
	    x[i] = -xOld + 2*xtmp - x[i+1];
            xOld = xtmp;
	    }
	x[i] = -xOld + 2*x[i] - x0;
	}

    /* output the data */
    for( int i=0; i<n; ++i ){
	int index=i;
        if( center )
	    index = (index+n/2)%n;
        output_it( x[index] );
	}

    return( 0 );
}

int ascii_in( complex &z ) {	// this is good for double AND complex input
    cin >> z;
    return( cin.good() );
    }

int fbits_in( complex &z ) {
    /* We assume that im=0. Our allocation method ensures this. */
    cin.read( (char *)&z, sizeof(double) );
    return( cin.good() );
    }

int zbits_in( complex &z ) {
    cin.read( (char *)&z, sizeof(complex) );
    return( cin.good() );
    }

void print_mag( const complex &x ) {
    cout << hypot( real(x), imag(x) ) << "\n";
}

void print_mag2( const complex &x ) {
    cout << real(x)*real(x) + imag(x)*imag(x) << "\n";
}

void print_double( const complex &x ) {
    cout << real(x) << "\n";
}

void print_cx( const complex &x ) {
    cout << x << "\n";
}

void write_mag( const complex &x ) {
    double mag = hypot( real(x), imag(x) );
    cout.write( (char *)(&mag), sizeof(double) );
    }

void write_mag2( const complex &x ) {
    double mag2 = real(x)*real(x) + imag(x)*imag(x);
    cout.write( (char *)(&mag2), sizeof(double) );
    }

void write_double( const complex &x ) {
    cout.write( (char *)(&x), sizeof(double) );
    }

void write_cx( const complex &x ) {
    cout.write( (char *)(&x), sizeof(complex) );
    }

/* read data and allocate a cx vector */
int get_cx_vector()
{
    complex input;

    n=1;
    x = (complex *) calloc( (unsigned)1, (unsigned) sizeof( complex ) );

    for( int count=0; get_input( input ); count++ ) {
    	if( count >= n ) {
    		n *= 2;
    		x = (complex *) realloc( (char *) x, (unsigned) n * sizeof( complex ) );
    		if( x == NULL ) {
    			fprintf( stderr, "unable to allocate sufficient storage \n-- terminating with %d samples read in.\n", n/2 );
    			return( -1 );
    			}
    		}
    	 x[count] = input;
	 }

    if( count < n )
	fprintf( stderr, "fft: appended %d zeroes to the input\n", n-count );
    if( debug ) /* echo the input */
        for( int i=0; i<n; ++i )
	    cout << x[i] << "\n";
    return( 0 );
}

/* Do an in-place decimation-in-time fft */
/* See Figure 6.10 of "Digital Signal Processing" by Oppenheim and Schafer. */
void fft()
{
    void reorder( complex*, int, int );
    void butterfly( complex&, complex& );

    register int i, j;
    int log2n=0;
    register int increment, offset;
    double theta;
    complex w, w0;

    for( int tmp=n; (tmp >>= 1) != 0 ; log2n++ ) /* n = 2^log2n */
	;
    reorder( x, n, log2n );

    /* do the fft */
    increment = 2; offset = 1; 
    while( log2n-- > 0 ) { 
    	/* this is the first iteration of every stage, 
    	   it has been separated out since w is one for the first itn */
    	for( i=0; i<n; i+=increment )
    		butterfly( x[i], x[i+offset] );

    	theta = M_PI / offset;
    	w0 = complex( cos(theta), -sin(theta) );
    	w = w0;
    	for( j=1; j<offset; ++j ) {
    		for( i = j + offset; i<n; i+=increment ) {
    			x[i] *= w;
    			butterfly( x[i-offset], x[i] );
    			}
    		w *= w0;
    		}

    	offset = increment;
    	increment *= 2;
    	}
}

void reorder( complex* x, int n, int m )
{ /* order the input vector by the bit-reversed indices */
    void cx_swap( complex&, complex& );
    int bit_reverse( int, int );

    for( int i=1; i<n; ++i ) {
    	int j = bit_reverse( i, m );
    	if( i<j )
    		cx_swap( x[i], x[j] );
    	}
}

int bit_reverse( int i, int m )
{
    int ir;
    ir = i&1;
    while( --m > 0 ) {
    	i = i>>1;
    	ir = ir<<1 | i&1;
    	}
    return( ir );
}

/* exchange two complex numbers */
void cx_swap( complex& x, complex& y )
{
    complex tmp;
    tmp = x;
    x = y;
    y = tmp;
}

/* do the basic butterfly operation */
void butterfly( complex &x, complex &y )
{
    complex xtmp;
    xtmp = x;
    x = xtmp + y;
    y = xtmp - y;
}
